# (C) Copyright 2007 Riverbank Computing Limited
# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The PyQt specific implementations the action manager internal classes. """


from inspect import getfullargspec


from pyface.qt import QtGui, QtCore


from traits.api import Any, Bool, HasTraits


from pyface.action.action_event import ActionEvent


class PyfaceWidgetAction(QtGui.QWidgetAction):
    def __init__(self, parent, action):
        super(PyfaceWidgetAction, self).__init__(parent)
        self.action = action

    def createWidget(self, parent):
        widget = self.action.create_control(parent)
        widget._action = self.action
        return widget


class _MenuItem(HasTraits):
    """ A menu item representation of an action item. """

    # '_MenuItem' interface ------------------------------------------------

    # Is the item checked?
    checked = Bool(False)

    # A controller object we delegate taking actions through (if any).
    controller = Any()

    # Is the item enabled?
    enabled = Bool(True)

    # Is the item visible?
    visible = Bool(True)

    # The radio group we are part of (None if the menu item is not part of such
    # a group).
    group = Any()

    # The toolkit control.
    control = Any()

    # The toolkit control id.
    control_id = None

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(self, parent, menu, item, controller):
        """ Creates a new menu item for an action item. """

        self.item = item
        action = item.action

        # FIXME v3: This is a wx'ism and should be hidden in the toolkit code.
        self.control_id = None

        if action.style == "widget":
            self.control = PyfaceWidgetAction(parent, action)
            menu.addAction(self.control)
        elif action.image is None:
            self.control = menu.addAction(
                action.name, self._qt4_on_triggered, action.accelerator
            )
        else:
            self.control = menu.addAction(
                action.image.create_icon(),
                action.name,
                self._qt4_on_triggered,
                action.accelerator,
            )
        menu.menu_items.append(self)

        self.control.setToolTip(action.tooltip)
        self.control.setWhatsThis(action.description)
        self.control.setEnabled(action.enabled)
        self.control.setVisible(action.visible)

        if getattr(action, "menu_role", False):
            if action.menu_role == "About":
                self.control.setMenuRole(QtGui.QAction.AboutRole)
            elif action.menu_role == "Preferences":
                self.control.setMenuRole(QtGui.QAction.PreferencesRole)

        if action.style == "toggle":
            self.control.setCheckable(True)
            self.control.setChecked(action.checked)
        elif action.style == "radio":
            # Create an action group if it hasn't already been done.
            try:
                ag = item.parent._qt4_ag
            except AttributeError:
                ag = item.parent._qt4_ag = QtGui.QActionGroup(parent)

            self.control.setActionGroup(ag)

            self.control.setCheckable(True)
            self.control.setChecked(action.checked)

        # Listen for trait changes on the action (so that we can update its
        # enabled/disabled/checked state etc).
        action.observe(self._on_action_enabled_changed, "enabled")
        action.observe(self._on_action_visible_changed, "visible")
        action.observe(self._on_action_checked_changed, "checked")
        action.observe(self._on_action_name_changed, "name")
        action.observe(self._on_action_accelerator_changed, "accelerator")
        action.observe(self._on_action_image_changed, "image")
        action.observe(self._on_action_tooltip_changed, "tooltip")

        # Detect if the control is destroyed.
        self.control.destroyed.connect(self._qt4_on_destroyed)

        if controller is not None:
            self.controller = controller
            controller.add_to_menu(self)

    def dispose(self):
        action = self.item.action
        action.observe(self._on_action_enabled_changed, "enabled", remove=True)
        action.observe(self._on_action_visible_changed, "visible", remove=True)
        action.observe(self._on_action_checked_changed, "checked", remove=True)
        action.observe(self._on_action_name_changed, "name", remove=True)
        action.observe(
            self._on_action_accelerator_changed, "accelerator", remove=True
        )
        action.observe(self._on_action_image_changed, "image", remove=True)
        action.observe(self._on_action_tooltip_changed, "tooltip", remove=True)

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _qt4_on_destroyed(self, control=None):
        """ Delete the reference to the control to avoid attempting to talk to
        it again.
        """
        self.control = None

    def _qt4_on_triggered(self):
        """ Called when the menu item has been clicked. """

        action = self.item.action
        action_event = ActionEvent()

        is_checkable = action.style in ["radio", "toggle"]

        # Perform the action!
        if self.controller is not None:
            if is_checkable:
                self.checked = action.checked = self.control.isChecked()

            # Most of the time, action's do no care about the event (it
            # contains information about the time the event occurred etc), so
            # we only pass it if the perform method requires it. This is also
            # useful as Traits UI controllers *never* require the event.
            argspec = getfullargspec(self.controller.perform)

            # If the only arguments are 'self' and 'action' then don't pass
            # the event!
            if len(argspec.args) == 2:
                self.controller.perform(action)

            else:
                self.controller.perform(action, action_event)

        else:
            if is_checkable:
                self.checked = action.checked = self.control.isChecked()

            # Most of the time, action's do no care about the event (it
            # contains information about the time the event occurred etc), so
            # we only pass it if the perform method requires it.
            argspec = getfullargspec(action.perform)

            # If the only argument is 'self' then don't pass the event!
            if len(argspec.args) == 1:
                action.perform()

            else:
                action.perform(action_event)

    # Trait event handlers -------------------------------------------------

    def _enabled_changed(self):
        """ Called when our 'enabled' trait is changed. """
        if self.control is not None:
            self.control.setEnabled(self.enabled)

    def _visible_changed(self):
        """ Called when our 'visible' trait is changed. """
        if self.control is not None:
            self.control.setVisible(self.visible)

    def _checked_changed(self):
        """ Called when our 'checked' trait is changed. """
        if self.control is not None:
            self.control.setChecked(self.checked)

    def _on_action_enabled_changed(self, event):
        """ Called when the enabled trait is changed on an action. """
        action = event.object
        if self.control is not None:
            self.control.setEnabled(action.enabled)

    def _on_action_visible_changed(self, event):
        """ Called when the visible trait is changed on an action. """
        action = event.object
        if self.control is not None:
            self.control.setVisible(action.visible)

    def _on_action_checked_changed(self, event):
        """ Called when the checked trait is changed on an action. """
        action = event.object
        if self.control is not None:
            self.control.setChecked(action.checked)

    def _on_action_name_changed(self, event):
        """ Called when the name trait is changed on an action. """
        action = event.object
        if self.control is not None:
            self.control.setText(action.name)

    def _on_action_accelerator_changed(self, event):
        """ Called when the accelerator trait is changed on an action. """
        action = event.object
        if self.control is not None:
            self.control.setShortcut(action.accelerator)

    def _on_action_image_changed(self, event):
        """ Called when the accelerator trait is changed on an action. """
        action = event.object
        if self.control is not None:
            self.control.setIcon(action.image.create_icon())

    def _on_action_tooltip_changed(self, event):
        """ Called when the accelerator trait is changed on an action. """
        action = event.object
        if self.control is not None:
            self.control.setToolTip(action.tooltip)


class _Tool(HasTraits):
    """ A tool bar tool representation of an action item. """

    # '_Tool' interface ----------------------------------------------------

    # Is the item checked?
    checked = Bool(False)

    # A controller object we delegate taking actions through (if any).
    controller = Any()

    # Is the item enabled?
    enabled = Bool(True)

    # Is the item visible?
    visible = Bool(True)

    # The radio group we are part of (None if the tool is not part of such a
    # group).
    group = Any()

    # The toolkit control.
    control = Any()

    # The toolkit control id.
    control_id = None

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(
        self, parent, tool_bar, image_cache, item, controller, show_labels
    ):
        """ Creates a new tool bar tool for an action item. """

        self.item = item
        self.tool_bar = tool_bar
        action = item.action

        if action.style == "widget":
            widget = action.create_control(tool_bar)
            self.control = tool_bar.addWidget(widget)
        elif action.image is None:
            self.control = tool_bar.addAction(action.name)
        else:
            size = tool_bar.iconSize()
            image = action.image.create_icon((size.width(), size.height()))
            self.control = tool_bar.addAction(image, action.name)
        tool_bar.tools.append(self)

        self.control.triggered.connect(self._qt4_on_triggered)

        self.control.setToolTip(action.tooltip)
        self.control.setWhatsThis(action.description)
        self.control.setEnabled(action.enabled)
        self.control.setVisible(action.visible)

        if action.style == "toggle":
            self.control.setCheckable(True)
            self.control.setChecked(action.checked)
        elif action.style == "radio":
            # Create an action group if it hasn't already been done.
            try:
                ag = item.parent._qt4_ag
            except AttributeError:
                ag = item.parent._qt4_ag = QtGui.QActionGroup(parent)

            self.control.setActionGroup(ag)

            self.control.setCheckable(True)
            self.control.setChecked(action.checked)

        # Keep a reference in the action.  This is done to make sure we live as
        # long as the action (and still respond to its signals) and don't die
        # if the manager that created us is garbage collected.
        self.control._tool_instance = self

        # Listen for trait changes on the action (so that we can update its
        # enabled/disabled/checked state etc).
        action.observe(self._on_action_enabled_changed, "enabled")
        action.observe(self._on_action_visible_changed, "visible")
        action.observe(self._on_action_checked_changed, "checked")
        action.observe(self._on_action_name_changed, "name")
        action.observe(self._on_action_accelerator_changed, "accelerator")
        action.observe(self._on_action_image_changed, "image")
        action.observe(self._on_action_tooltip_changed, "tooltip")

        # Detect if the control is destroyed.
        self.control.destroyed.connect(self._qt4_on_destroyed)

        if controller is not None:
            self.controller = controller
            controller.add_to_toolbar(self)

    def dispose(self):
        action = self.item.action
        action.observe(self._on_action_enabled_changed, "enabled", remove=True)
        action.observe(self._on_action_visible_changed, "visible", remove=True)
        action.observe(self._on_action_checked_changed, "checked", remove=True)
        action.observe(self._on_action_name_changed, "name", remove=True)
        action.observe(
            self._on_action_accelerator_changed, "accelerator", remove=True
        )
        action.observe(self._on_action_image_changed, "image", remove=True)
        action.observe(self._on_action_tooltip_changed, "tooltip", remove=True)

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _qt4_on_destroyed(self, control=None):
        """ Delete the reference to the control to avoid attempting to talk to
        it again.
        """
        self.control = None

    def _qt4_on_triggered(self):
        """ Called when the tool bar tool is clicked. """

        action = self.item.action
        action_event = ActionEvent()

        # Perform the action!
        if self.controller is not None:
            self.checked = action.checked = self.control.isChecked()

            # Most of the time, action's do no care about the event (it
            # contains information about the time the event occurred etc), so
            # we only pass it if the perform method requires it. This is also
            # useful as Traits UI controllers *never* require the event.
            argspec = getfullargspec(self.controller.perform)

            # If the only arguments are 'self' and 'action' then don't pass
            # the event!
            if len(argspec.args) == 2:
                self.controller.perform(action)

            else:
                self.controller.perform(action, action_event)

        else:
            self.checked = action.checked = self.control.isChecked()

            # Most of the time, action's do no care about the event (it
            # contains information about the time the event occurred etc), so
            # we only pass it if the perform method requires it.
            argspec = getfullargspec(action.perform)

            # If the only argument is 'self' then don't pass the event!
            if len(argspec.args) == 1:
                action.perform()

            else:
                action.perform(action_event)

    # Trait event handlers -------------------------------------------------

    def _enabled_changed(self):
        """ Called when our 'enabled' trait is changed. """
        if self.control is not None:
            self.control.setEnabled(self.enabled)

    def _visible_changed(self):
        """ Called when our 'visible' trait is changed. """
        if self.control is not None:
            self.control.setVisible(self.visible)

    def _checked_changed(self):
        """ Called when our 'checked' trait is changed. """
        if self.control is not None:
            self.control.setChecked(self.checked)

    def _on_action_enabled_changed(self, event):
        """ Called when the enabled trait is changed on an action. """
        action = event.object
        if self.control is not None:
            self.control.setEnabled(action.enabled)

    def _on_action_visible_changed(self, event):
        """ Called when the visible trait is changed on an action. """
        action = event.object
        if self.control is not None:
            self.control.setVisible(action.visible)

    def _on_action_checked_changed(self, event):
        """ Called when the checked trait is changed on an action. """
        action = event.object
        if self.control is not None:
            self.control.setChecked(action.checked)

    def _on_action_name_changed(self, event):
        """ Called when the name trait is changed on an action. """
        action = event.object
        if self.control is not None:
            self.control.setText(action.name)

    def _on_action_accelerator_changed(self, event):
        """ Called when the accelerator trait is changed on an action. """
        action = event.object
        if self.control is not None:
            self.control.setShortcut(action.accelerator)

    def _on_action_image_changed(self, event):
        """ Called when the accelerator trait is changed on an action. """
        action = event.object
        if self.control is not None:
            size = self.tool_bar.iconSize()
            self.control.setIcon(
                action.image.create_icon((size.width(), size.height()))
            )

    def _on_action_tooltip_changed(self, event):
        """ Called when the accelerator trait is changed on an action. """
        action = event.object
        if self.control is not None:
            self.control.setToolTip(action.tooltip)


class _PaletteTool(HasTraits):
    """ A tool palette representation of an action item. """

    # '_PaletteTool' interface ---------------------------------------------

    # The radio group we are part of (None if the tool is not part of such a
    # group).
    group = Any()

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(self, tool_palette, image_cache, item, show_labels):
        """ Creates a new tool palette tool for an action item. """

        self.item = item
        self.tool_palette = tool_palette

        action = self.item.action
        label = action.name

        if action.style == "widget":
            raise NotImplementedError(
                "Qt does not support widgets in palettes"
            )

        # Tool palette tools never have '...' at the end.
        if label.endswith("..."):
            label = label[:-3]

        # And they never contain shortcuts.
        label = label.replace("&", "")

        image = action.image.create_image()
        path = action.image.absolute_path
        bmp = image_cache.get_bitmap(path)

        kind = action.style
        tooltip = action.tooltip
        longtip = action.description

        if not show_labels:
            label = ""

        # Add the tool to the tool palette.
        self.tool_id = tool_palette.add_tool(
            label, bmp, kind, tooltip, longtip
        )
        tool_palette.toggle_tool(self.tool_id, action.checked)
        tool_palette.enable_tool(self.tool_id, action.enabled)
        tool_palette.on_tool_event(self.tool_id, self._on_tool)

        # Listen to the trait changes on the action (so that we can update its
        # enabled/disabled/checked state etc).
        action.observe(self._on_action_enabled_changed, "enabled")
        action.observe(self._on_action_checked_changed, "checked")

        return

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    # Trait event handlers -------------------------------------------------

    def _on_action_enabled_changed(self, event):
        """ Called when the enabled trait is changed on an action. """
        action = event.object
        self.tool_palette.enable_tool(self.tool_id, action.enabled)

    def _on_action_checked_changed(self, event):
        """ Called when the checked trait is changed on an action. """
        action = event.object
        if action.style == "radio":
            # If we're turning this one on, then we need to turn all the others
            # off.  But if we're turning this one off, don't worry about the
            # others.
            if event.new:
                for item in self.item.parent.items:
                    if item is not self.item:
                        item.action.checked = False

        # This will *not* emit a tool event.
        self.tool_palette.toggle_tool(self.tool_id, event.new)

        return

    # Tool palette event handlers -----------------------------------------#

    def _on_tool(self, event):
        """ Called when the tool palette button is clicked. """

        action = self.item.action
        action_event = ActionEvent()

        is_checkable = action.style == "radio" or action.style == "check"

        # Perform the action!
        action.checked = self.tool_palette.get_tool_state(self.tool_id) == 1
        action.perform(action_event)

        return

    def dispose(self):
        action = self.item.action
        action.observe(self._on_action_enabled_changed, "enabled", remove=True)
        action.observe(self._on_action_checked_changed, "checked", remove=True)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from wx import CustomDataObject, DataFormat, DataObject, DataObjectComposite

from traits.api import Instance, provides

from pyface.data_view.i_data_wrapper import IDataWrapper, MDataWrapper


@provides(IDataWrapper)
class DataWrapper(MDataWrapper):
    """ WxPython implementaton of IDataWrapper.

    This wraps a DataObjectComposite which is assumed to contain a collection
    of CustomDataObjects that store data associated by mimetype.  Any other
    DataObjects in the DataObjectComposite are ignored.
    """

    #: We always have a a composite data object with custom data objects in it
    toolkit_data = Instance(
        DataObjectComposite,
        args=(),
        allow_none=False,
    )

    def mimetypes(self):
        """ Return a set of mimetypes holding data.

        Returns
        -------
        mimetypes : set of str
            The set of mimetypes currently storing data in the toolkit data
            object.
        """
        return {
            wx_format.GetId()
            for wx_format in self.toolkit_data.GetAllFormats()
        }

    def get_mimedata(self, mimetype):
        """ Get raw data for the given media type.

        Parameters
        ----------
        mimetype : str
            The mime media type to be extracted.

        Returns
        -------
        mimedata : bytes
            The mime media data as bytes.
        """
        wx_format = DataFormat(mimetype)
        if self.toolkit_data.IsSupported(wx_format):
            data_object = self.toolkit_data.GetObject(wx_format)
            if isinstance(data_object, CustomDataObject):
                return bytes(data_object.GetData())
        return None

    def set_mimedata(self, mimetype, raw_data):
        """ Set raw data for the given media type.

        Parameters
        ----------
        mimetype : str
            The mime media type to be extracted.
        mimedata : bytes
            The mime media data encoded as bytes..
        """
        wx_format = DataFormat(mimetype)
        if self.toolkit_data.IsSupported(wx_format, dir=DataObject.Set):
            data_object = self.toolkit_data.GetObject(wx_format)
        else:
            data_object = CustomDataObject(wx_format)
            self.toolkit_data.Add(data_object)
        data_object.SetData(raw_data)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from pyface.qt import QtCore

# ----------------------------------------------------------------------------
# Functions.
# ----------------------------------------------------------------------------


def set_focus(control):
    """ Assign keyboard focus to the given control.

    Ideally, we would just call ``setFocus()`` on the control and let Qt do the
    right thing. Unfortunately, this method is implemented in the most naive
    manner possible, and is essentially a no-op if the toplevel widget does not
    itself accept focus. We adopt the following procedure:

    1. If the control itself accepts focus, use it. This is important since the
       control may have custom focus dispatching logic.

    2. Otherwise, if there is a child widget of the control that previously had
       focus, use it.

    3. Finally, have Qt determine the next item using its internal logic. Qt
       will only restrict itself to this widget's children if it is a Qt::Window
       or Qt::SubWindow, hence the hack below.
    """
    if control.focusPolicy() != QtCore.Qt.NoFocus:
        control.setFocus()
    else:
        widget = control.focusWidget()
        if widget:
            widget.setFocus()
        else:
            flags = control.windowFlags()
            control.setWindowFlags(flags | QtCore.Qt.SubWindow)
            try:
                control.focusNextChild()
            finally:
                control.setWindowFlags(flags)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
"""
Index Managers
==============

This module provides a number of classes for efficiently managing the
mapping between different ways of representing indices.  To do so, each
index manager provides an intermediate, opaque index object that is
suitable for use in these situations and is guaranteed to have a long
enough life that it will not change or be garbage collected while a C++
object has a reference to it.

The wx DataView classes expect to be given an integer id value that is
stable and can be used to return the parent reference id.

And Qt's ModelView system expects to be given a pointer to an object
that is long-lived (in particular, it will not be garbage-collected
during the lifetime of a QModelIndex) and which can be used to find
the parent object of the current object.

The default representation of an index from the point of view of the
data view infrastructure is a sequence of integers, giving the index at
each level of the hierarchy.  DataViewModel classes can then use these
indices to identify objects in the underlying data model.

There are three main classes defined in the module: AbstractIndexManager,
IntIndexManager, and TupleIndexManager.

AbstractIndexManager
    An ABC that defines the API

IntIndexManager
    An efficient index manager for non-hierarchical data, such as
    lists, tables and 2D arrays.

TupleIndexManager
    An index manager that handles non-hierarchical data while trying
    to be fast and memory efficient.

The two concrete subclasses should be sufficient for most cases, but advanced
users may create their own if for some reason the provided managers do not
work well for a particular situation.  Developers who implement this API
need to be mindful of the requirements on the lifetime and identity
constraints required by the various toolkit APIs.

"""

from abc import abstractmethod

from traits.api import ABCHasStrictTraits, Dict, Int, Tuple


#: The singular root object for all index managers.
Root = ()


class AbstractIndexManager(ABCHasStrictTraits):
    """ Abstract base class for index managers.
    """

    @abstractmethod
    def create_index(self, parent, row):
        """ Given a parent index and a row number, create an index.

        The internal structure of the index should not matter to
        consuming code.  However obejcts returned from this method
        should persist until the reset method is called.

        Parameters
        ----------
        parent : index object
            The parent index object.
        row : int
            The position of the resuling index in the parent's children.

        Returns
        -------
        index : index object
            The resulting opaque index object.

        Raises
        ------
        IndexError
            Negative row values raise an IndexError exception.
        RuntimeError
            If asked to create a persistent index for a parent and row
            where that is not possible, a RuntimeError will be raised.
        """
        raise NotImplementedError()

    @abstractmethod
    def get_parent_and_row(self, index):
        """ Given an index object, return the parent index and row.

        Parameters
        ----------
        index : index object
            The opaque index object.

        Returns
        -------
        parent : index object
            The parent index object.
        row : int
            The position of the resuling index in the parent's children.

        Raises
        ------
        IndexError
            If the Root object is passed as the index, this method will
            raise an IndexError, as it has no parent.
        """
        raise NotImplementedError()

    def from_sequence(self, indices):
        """ Given a sequence of indices, return the index object.

        The default implementation starts at the root and repeatedly calls
        create_index() to find the index at each level, returning the final
        value.

        Parameters
        ----------
        indices : sequence of int
            The row location at each level of the hierarchy.

        Returns
        -------
        index : index object
            The persistent index object associated with this sequence.

        Raises
        ------
        RuntimeError
            If asked to create a persistent index for a sequence of indices
            where that is not possible, a RuntimeError will be raised.
        """
        index = Root
        for row in indices:
            index = self.create_index(index, row)
        return index

    def to_sequence(self, index):
        """ Given an index, return the corresponding sequence of row values.

        The default implementation repeatedly calls get_parent_and_row()
        to walk up the hierarchy and push the row values into the start
        of the sequence.

        Parameters
        ----------
        index : index object
            The opaque index object.

        Returns
        -------
        sequence : tuple of int
            The row location at each level of the hierarchy.
        """
        result = ()
        while index != Root:
            index, row = self.get_parent_and_row(index)
            result = (row,) + result
        return result

    @abstractmethod
    def from_id(self, id):
        """ Given an integer id, return the corresponding index.

        Parameters
        ----------
        id : int
            An integer object id value.

        Returns
        -------
        index : index object
            The persistent index object associated with this id.
        """
        raise NotImplementedError()

    @abstractmethod
    def id(self, index):
        """ Given an index, return the corresponding id.

        Parameters
        ----------
        index : index object
            The persistent index object.

        Returns
        -------
        id : int
            The associated integer object id value.
        """
        raise NotImplementedError()

    def reset(self):
        """ Reset any caches and other state.

        Resettable traits in subclasses are indicated by having
        ``can_reset=True`` metadata.  This is provided to allow
        toolkit code to clear caches to prevent memory leaks when
        working with very large tables.

        Care should be taken when calling this method, as Qt may
        crash if a QModelIndex is referencing an index that no
        longer has a reference in a cache.

        For some IndexManagers, particularly for those which are flat
        or static, reset() may do nothing.
        """
        resettable_traits = self.trait_names(can_reset=True)
        self.reset_traits(resettable_traits)


class IntIndexManager(AbstractIndexManager):
    """ Efficient IndexManager for non-hierarchical indexes.

    This is a simple index manager for flat data structures.  The
    index values returned are either the Root, or simple integers
    that indicate the position of the index as a child of the root.

    While it cannot handle nested data, this index manager can
    operate without having to perform any caching, and so is very
    efficient.
    """

    def create_index(self, parent, row):
        """ Given a parent index and a row number, create an index.

        This should only ever be called with Root as the parent.

        Parameters
        ----------
        parent : index object
            The parent index object.
        row : non-negative int
            The position of the resulting index in the parent's children.

        Returns
        -------
        index : index object
            The resulting opaque index object.

        Raises
        ------
        IndexError
            Negative row values raise an IndexError exception.
        RuntimeError
            If the parent is not the Root, a RuntimeError will be raised
        """
        if row < 0:
            raise IndexError("Row must be non-negative.  Got {}".format(row))
        if parent != Root:
            raise RuntimeError(
                "{} cannot create persistent index value for {}.".format(
                    self.__class__.__name__,
                    (parent, row)
                )
            )
        return row

    def get_parent_and_row(self, index):
        """ Given an index object, return the parent index and row.

        Parameters
        ----------
        index : index object
            The opaque index object.

        Returns
        -------
        parent : index object
            The parent index object.
        row : int
            The position of the resuling index in the parent's children.

        Raises
        ------
        IndexError
            If the Root object is passed as the index, this method will
            raise an IndexError, as it has no parent.
        """
        if index == Root:
            raise IndexError("Root index has no parent.")
        return Root, int(index)

    def from_id(self, id):
        """ Given an integer id, return the corresponding index.

        Parameters
        ----------
        id : int
            An integer object id value.

        Returns
        -------
        index : index object
            The persistent index object associated with this id.
        """
        if id == 0:
            return Root
        return id - 1

    def id(self, index):
        """ Given an index, return the corresponding id.

        Parameters
        ----------
        index : index object
            The persistent index object.

        Returns
        -------
        id : int
            The associated integer object id value.
        """
        if index == Root:
            return 0
        return index + 1


class TupleIndexManager(AbstractIndexManager):

    #: A dictionary that maps tuples to the canonical version of the tuple.
    _cache = Dict(Tuple, Tuple, {Root: Root}, can_reset=True)

    #: A dictionary that maps ids to the canonical version of the tuple.
    _id_cache = Dict(Int, Tuple, {0: Root}, can_reset=True)

    def create_index(self, parent, row):
        """ Given a parent index and a row number, create an index.

        Parameters
        ----------
        parent : index object
            The parent index object.
        row : non-negative int
            The position of the resulting index in the parent's children.

        Returns
        -------
        index : index object
            The resulting opaque index object.

        Raises
        ------
        IndexError
            Negative row values raise an IndexError exception.
        """
        if row < 0:
            raise IndexError("Row must be non-negative.  Got {}".format(row))

        index = (parent, row)
        canonical_index = self._cache.setdefault(index, index)
        self._id_cache[self.id(canonical_index)] = canonical_index
        return canonical_index

    def get_parent_and_row(self, index):
        """ Given an index object, return the parent index and row.

        Parameters
        ----------
        index : index object
            The opaque index object.

        Returns
        -------
        parent : index object
            The parent index object.
        row : int
            The position of the resuling index in the parent's children.

        Raises
        ------
        IndexError
            If the Root object is passed as the index, this method will
            raise an IndexError, as it has no parent.
        """
        if index == Root:
            raise IndexError("Root index has no parent.")
        return index

    def from_id(self, id):
        """ Given an integer id, return the corresponding index.

        Parameters
        ----------
        id : int
            An integer object id value.

        Returns
        -------
        index : index object
            The persistent index object associated with this id.
        """
        return self._id_cache[id]

    def id(self, index):
        """ Given an index, return the corresponding id.

        Parameters
        ----------
        index : index object
            The persistent index object.

        Returns
        -------
        id : int
            The associated integer object id value.
        """
        if index == Root:
            return 0
        canonical_index = self._cache.setdefault(index, index)
        return id(canonical_index)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from unittest import TestCase
from unittest.mock import Mock

from traits.testing.api import UnittestTools

from pyface.color import Color
from pyface.data_view.value_types.constant_value import ConstantValue
from pyface.image_resource import ImageResource


class TestConstantValue(UnittestTools, TestCase):

    def setUp(self):
        self.model = Mock()

    def test_defaults(self):
        value_type = ConstantValue()
        self.assertEqual(value_type.text, "")
        self.assertEqual(value_type.tooltip, "")

    def test_has_editor_value(self):
        value_type = ConstantValue()
        self.assertFalse(value_type.has_editor_value(self.model, [0], [0]))

    def test_has_text(self):
        value_type = ConstantValue()
        self.assertFalse(value_type.has_text(self.model, [0], [0]))

    def test_has_text_true(self):
        value_type = ConstantValue(text="something")
        self.assertTrue(value_type.has_text(self.model, [0], [0]))

    def test_get_text(self):
        value_type = ConstantValue(text="something")
        self.assertEqual(
            value_type.get_text(self.model, [0], [0]),
            "something"
        )

    def test_text_changed(self):
        value_type = ConstantValue()
        with self.assertTraitChanges(value_type, 'updated'):
            value_type.text = 'something'
        self.assertEqual(value_type.text, 'something')

    def test_has_color_default(self):
        value_type = ConstantValue()
        self.assertFalse(value_type.has_color(self.model, [0], [0]))

    def test_has_color(self):
        value_type = ConstantValue(color=Color(rgba=(0.4, 0.2, 0.6, 0.8)))
        self.assertTrue(value_type.has_color(self.model, [0], [0]))

    def test_get_color_default(self):
        value_type = ConstantValue()
        self.assertIsNone(value_type.get_color(self.model, [0], [0]))

    def test_get_color(self):
        value_type = ConstantValue(color='rebeccapurple')
        self.assertEqual(
            value_type.get_color(self.model, [0], [0]),
            Color(rgba=(0.4, 0.2, 0.6, 1.0))
        )

    def test_get_color_changed(self):
        value_type = ConstantValue()
        with self.assertTraitChanges(value_type, 'updated'):
            value_type.color = Color(rgba=(0.4, 0.2, 0.6, 0.8))
        self.assertEqual(
            value_type.get_color(self.model, [0], [0]),
            Color(rgba=(0.4, 0.2, 0.6, 0.8))
        )

    def test_get_color_rgba_changed(self):
        value_type = ConstantValue(color=Color())
        with self.assertTraitChanges(value_type, 'updated'):
            value_type.color.rgba = (0.4, 0.2, 0.6, 0.8)
        self.assertEqual(
            value_type.get_color(self.model, [0], [0]),
            Color(rgba=(0.4, 0.2, 0.6, 0.8))
        )

    def test_has_image(self):
        value_type = ConstantValue()
        self.assertFalse(value_type.has_image(self.model, [0], [0]))

    def test_has_image_true(self):
        value_type = ConstantValue(image="question")
        self.assertTrue(value_type.has_image(self.model, [0], [0]))

    def test_get_image(self):
        image = ImageResource("question")
        value_type = ConstantValue(image=image)
        self.assertEqual(
            value_type.get_image(self.model, [0], [0]),
            image
        )

    def test_get_image_none(self):
        value_type = ConstantValue()
        image = value_type.get_image(self.model, [0], [0])
        self.assertEqual(image.name, "image_not_found")

    def test_image_changed(self):
        value_type = ConstantValue()
        image = ImageResource("question")
        with self.assertTraitChanges(value_type, 'updated'):
            value_type.image = image
        self.assertEqual(value_type.image, image)

    def test_has_tooltip(self):
        value_type = ConstantValue()
        self.assertFalse(value_type.has_tooltip(self.model, [0], [0]))

    def test_has_tooltip_true(self):
        value_type = ConstantValue(tooltip="something")
        self.assertTrue(value_type.has_tooltip(self.model, [0], [0]))

    def test_get_tooltip(self):
        value_type = ConstantValue(tooltip="something")
        self.assertEqual(
            value_type.get_tooltip(self.model, [0], [0]),
            "something"
        )

    def test_tooltip_changed(self):
        value_type = ConstantValue()
        with self.assertTraitChanges(value_type, 'updated'):
            value_type.tooltip = 'something'
        self.assertEqual(value_type.tooltip, 'something')



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import locale
from math import inf

from traits.api import Callable, Float

from pyface.data_view.data_view_errors import DataViewSetError
from .editable_value import EditableValue


def format_locale(value):
    return "{:n}".format(value)


class NumericValue(EditableValue):
    """ Data channels for a numeric value.
    """

    #: The minimum value for the numeric value.
    minimum = Float(-inf)

    #: The maximum value for the numeric value.
    maximum = Float(inf)

    #: A function that converts to the numeric type.
    evaluate = Callable()

    #: A function that converts the required type to a string for display.
    format = Callable(format_locale, update_value_type=True)

    #: A function that converts the required type from a display string.
    unformat = Callable(locale.delocalize)

    def is_valid(self, model, row, column, value):
        """ Whether or not the value within the specified range.

        Parameters
        ----------
        model : AbstractDataModel
            The data model holding the data.
        row : sequence of int
            The row in the data model being queried.
        column : sequence of int
            The column in the data model being queried.
        value : any
            The value to validate.

        Returns
        -------
        is_valid : bool
            Whether or not the value is valid.
        """
        try:
            return self.minimum <= value <= self.maximum
        except Exception:
            return False

    def get_editor_value(self, model, row, column):
        """ Get the numerical value for the editor to use.

        This uses the evaluate method to convert the underlying value to a
        number.

        Parameters
        ----------
        model : AbstractDataModel
            The data model holding the data.
        row : sequence of int
            The row in the data model being queried.
        column : sequence of int
            The column in the data model being queried.

        Returns
        -------
        editor_value : number
            Whether or not the value is editable.
        """
        # evaluate is needed to convert numpy types to python types so
        # Qt recognises them
        return self.evaluate(model.get_value(row, column))

    def get_text(self, model, row, column):
        """ Get the display text from the underlying value.

        Parameters
        ----------
        model : AbstractDataModel
            The data model holding the data.
        row : sequence of int
            The row in the data model being queried.
        column : sequence of int
            The column in the data model being queried.

        Returns
        -------
        text : str
            The text to display.
        """
        return self.format(model.get_value(row, column))

    def set_text(self, model, row, column, text):
        """ Set the text of the underlying value.

        Parameters
        ----------
        model : AbstractDataModel
            The data model holding the data.
        row : sequence of int
            The row in the data model being queried.
        column : sequence of int
            The column in the data model being queried.
        text : str
            The text to set.

        Raises
        -------
        DataViewSetError
            If the value cannot be set.
        """
        try:
            value = self.evaluate(self.unformat(text))
        except ValueError:
            raise DataViewSetError(
                "Can't evaluate value: {!r}".format(text)
            )
        self.set_editor_value(model, row, column, value)


class IntValue(NumericValue):
    """ Data channels for an integer value.
    """

    evaluate = Callable(int)


class FloatValue(NumericValue):
    """ Data channels for a floating point value.
    """

    evaluate = Callable(float)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""

API for the ``pyface.tasks`` submodule.

Tasks-specific Interfaces
-------------------------
- :class:`~.IDockPane`
- :class:`~.IEditor`
- :class:`~.IEditorAreaPane`
- :class:`~.ITaskPane`

Tasks, Tasks Application and related classes
--------------------------------------------

- :class:`~.AdvancedEditorAreaPane`
- :class:`~.DockPane`
- :class:`~.Editor`
- :class:`~.EditorAreaPane`
- :class:`~.SplitEditorAreaPane`
- :class:`~.Task`
- :class:`~.TasksApplication`
- :class:`~.TaskFactory`
- :class:`~.TaskPane`
- :class:`~.TaskWindow`

Tasks layout
------------
- :class:`~.TaskLayout`
- :class:`~.TaskWindowLayout`
- :class:`~.PaneItem`
- :class:`~.Tabbed`
- :class:`~.Splitter`
- :class:`~.HSplitter`
- :class:`~.VSplitter`

Traits-specific Tasks classes
-----------------------------
- :class:`~.TraitsDockPane`
- :class:`~.TraitsEditor`
- :class:`~.TraitsTaskPane`

Enaml-specific Tasks functionality
----------------------------------
- :class:`~.EnamlDockPane`
- :class:`~.EnamlEditor`
- :class:`~.EnamlTaskPane`

"""

from .advanced_editor_area_pane import AdvancedEditorAreaPane
from .split_editor_area_pane import SplitEditorAreaPane
from .dock_pane import DockPane
from .editor import Editor
from .editor_area_pane import EditorAreaPane
from .enaml_dock_pane import EnamlDockPane
from .enaml_editor import EnamlEditor
from .enaml_task_pane import EnamlTaskPane
from .i_dock_pane import IDockPane
from .i_editor import IEditor
from .i_editor_area_pane import IEditorAreaPane
from .i_task_pane import ITaskPane
from .task import Task
from .tasks_application import TasksApplication, TaskFactory
from .task_layout import (
    TaskLayout,
    PaneItem,
    Tabbed,
    Splitter,
    HSplitter,
    VSplitter,
)
from .task_pane import TaskPane
from .task_window import TaskWindow
from .task_window_layout import TaskWindowLayout
from .traits_dock_pane import TraitsDockPane
from .traits_editor import TraitsEditor
from .traits_task_pane import TraitsTaskPane



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" A row-oriented data model implementation.

This module provides a concrete implementation of a data model for the
case of non-hierarchical, row-oriented data.
"""
from collections.abc import Sequence

from traits.api import Instance, List, observe
from traits.observation.api import trait

from pyface.data_view.abstract_data_model import (
    AbstractDataModel, DataViewSetError
)
from pyface.data_view.index_manager import IntIndexManager
from pyface.data_view.data_models.data_accessors import AbstractDataAccessor


class RowTableDataModel(AbstractDataModel):
    """ A data model that presents a sequence of objects as rows.

    The data is expected to be a sequence of row objects, each object
    providing values for the columns via an AbstractDataAccessor subclass.
    Concrete implementations can be found in the data_accessors module that
    get data from attributes, indices of sequences, and keys of mappings,
    but for more complex situations, custom accessors can be defined.
    """

    #: A sequence of objects to display as rows.
    data = Instance(Sequence, allow_none=False)

    #: An object which describes how to map data for the row headers.
    row_header_data = Instance(AbstractDataAccessor, allow_none=False)

    #: An object which describes how to map data for each column.
    column_data = List(Instance(AbstractDataAccessor, allow_none=False))

    #: The index manager that helps convert toolkit indices to data view
    #: indices.
    index_manager = Instance(IntIndexManager, args=(), allow_none=False)

    # Data structure methods

    def get_column_count(self):
        """ How many columns in the data view model.

        Returns
        -------
        column_count : non-negative int
            The number of columns that the data view provides.
        """
        return len(self.column_data)

    def can_have_children(self, row):
        """ Whether or not a row can have child rows.

        Only the root has children.

        Parameters
        ----------
        row : sequence of int
            The indices of the row as a sequence from root to leaf.

        Returns
        -------
        can_have_children : bool
            Whether or not the row can ever have child rows.
        """
        return len(row) == 0

    def get_row_count(self, row):
        """ How many child rows the row currently has.

        Parameters
        ----------
        row : sequence of int
            The indices of the row as a sequence from root to leaf.

        Returns
        -------
        row_count : non-negative int
            The number of child rows that the row has.
        """
        if len(row) == 0:
            return len(self.data)
        else:
            return 0

    # Data value methods

    def get_value(self, row, column):
        """ Return the Python value for the row and column.

        This uses the row_header_data and column_data accessors to extract
        values for the row and column.

        Parameters
        ----------
        row : sequence of int
            The indices of the row as a sequence from root to leaf.
        column : sequence of int
            The indices of the column as a sequence of length 0 or 1.

        Returns
        -------
        value : any
            The value represented by the given row and column.
        """
        if len(column) == 0:
            column_data = self.row_header_data
        else:
            column_data = self.column_data[column[0]]
        if len(row) == 0:
            return column_data.title
        obj = self.data[row[0]]
        return column_data.get_value(obj)

    def can_set_value(self, row, column):
        """ Whether the value in the indicated row and column can be set.

        This uses the row_header_data and column_data accessors to determine
        if the value may be changed.

        Parameters
        ----------
        row : sequence of int
            The indices of the row as a sequence from root to leaf.
        column : sequence of int
            The indices of the column as a sequence of length 0 or 1.

        Returns
        -------
        can_set_value : bool
            Whether or not the value can be set.
        """
        if len(row) == 0:
            return False
        if len(column) == 0:
            column_data = self.row_header_data
        else:
            column_data = self.column_data[column[0]]
        obj = self.data[row[0]]
        return column_data.can_set_value(obj)

    def set_value(self, row, column, value):
        """ Set the Python value for the row and column.

        This uses the row_header_data and column_data accessors to set
        the value.

        Parameters
        ----------
        row : sequence of int
            The indices of the row as a sequence from root to leaf.
        column : sequence of int
            The indices of the column as a sequence of length 0 or 1.
        value : any
            The new value for the given row and column.

        Raises
        -------
        DataViewSetError
            If the value cannot be set.
        """
        if len(row) == 0:
            raise DataViewSetError("Can't set column titles.")
        if len(column) == 0:
            column_data = self.row_header_data
        else:
            column_data = self.column_data[column[0]]
        obj = self.data[row[0]]
        column_data.set_value(obj, value)
        self.values_changed = (row, column, row, column)

    def get_value_type(self, row, column):
        """ Return the value type of the given row and column.

        This uses the row_header_data and column_data accessors to get
        the value type.

        Parameters
        ----------
        row : sequence of int
            The indices of the row as a sequence from root to leaf.
        column : sequence of int
            The indices of the column as a sequence of length 0 or 1.

        Returns
        -------
        value_type : AbstractValueType or None
            The value type of the given row and column, or None if no value
            should be displayed.
        """
        if len(column) == 0:
            column_data = self.row_header_data
        else:
            column_data = self.column_data[column[0]]
        if len(row) == 0:
            return column_data.title_type
        return column_data.value_type

    # data update methods

    @observe("data")
    def _update_data(self, event):
        self.structure_changed = True

    @observe(trait("data", notify=False).list_items(optional=True))
    def _update_data_items(self, event):
        if len(event.added) != len(event.removed):
            # number of rows has changed
            self.structure_changed = True
        else:
            if isinstance(event.index, int):
                start = event.index
                stop = min(event.index + len(event.added), len(self.data)) - 1
            else:
                start = event.index.start
                stop = min(event.index.stop, len(self.data)) - 1
            self.values_changed = ((start,), (), (stop,), ())

    @observe('row_header_data')
    def _update_row_header_data(self, event):
        self.values_changed = ((), (), (), ())

    @observe('row_header_data:updated')
    def _update_row_header_data_event(self, event):
        if event.new[1] == 'value':
            if len(self.data) > 0:
                self.values_changed = ((0,), (), (len(self.data) - 1,), ())
        else:
            self.values_changed = ((), (), (), ())

    @observe('column_data.items')
    def _update_all_column_data_items(self, event):
        self.structure_changed = True

    @observe('column_data:items:updated')
    def _update_column_data(self, event):
        index = self.column_data.index(event.new[0])
        if event.new[1] == 'value':
            if len(self.data) > 0:
                self.values_changed = (
                    (0,), (index,), (len(self.data) - 1,), (index,)
                )
        else:
            self.values_changed = ((), (index,), (), (index,))

    # default data value

    def _data_default(self):
        return []



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A tree control with extensible node types. """


# avoid deprecation warning
from inspect import getfullargspec


from pyface.action.api import ActionEvent
from traits.api import Instance, List, Property


from .node_manager import NodeManager
from .node_type import NodeType
from .node_tree_model import NodeTreeModel
from .tree import Tree


class NodeTree(Tree):
    """ A tree control with extensible node types. """

    # 'Tree' interface -----------------------------------------------------

    # The model that provides the data for the tree.
    model = Instance(NodeTreeModel, ())

    # 'NodeTree' interface -------------------------------------------------

    # The node manager looks after all node types.
    node_manager = Property(Instance(NodeManager))

    # The node types in the tree.
    node_types = Property(List(NodeType))

    # ------------------------------------------------------------------------
    # 'NodeTree' interface.
    # ------------------------------------------------------------------------

    # Properties -----------------------------------------------------------

    # node_manager
    def _get_node_manager(self):
        """ Returns the root node of the tree. """

        return self.model.node_manager

    def _set_node_manager(self, node_manager):
        """ Sets the root node of the tree. """

        self.model.node_manager = node_manager

        return

    # node_types
    def _get_node_types(self):
        """ Returns the node types in the tree. """

        return self.model.node_manager.node_types

    def _set_node_types(self, node_types):
        """ Sets the node types in the tree. """

        self.model.node_manager.node_types = node_types

        return

    # ------------------------------------------------------------------------
    # 'Tree' interface.
    # ------------------------------------------------------------------------

    # Trait event handlers -------------------------------------------------

    def _node_activated_changed(self, obj):
        """ Called when a node has been activated (i.e., double-clicked). """

        default_action = self.model.get_default_action(obj)
        if default_action is not None:
            self._perform_default_action(default_action, obj)

    def _node_right_clicked_changed(self, obj_point):
        """ Called when the right mouse button is clicked on the tree. """

        obj, point = obj_point
        # Add the node that the right-click occurred on to the selection.
        self.select(obj)

        # fixme: This is a hack to allow us to attach the node that the
        # right-clicked occurred on to the action event.
        self._context = obj

        # Ask the model for the node's context menu.
        menu_manager = self.model.get_context_menu(obj)
        if menu_manager is not None:
            self._popup_menu(menu_manager, obj, point)

        return

    # ------------------------------------------------------------------------
    # 'ActionController' interface.
    # ------------------------------------------------------------------------

    def add_to_menu(self, menu_item):
        """ Adds a menu item to a menu bar. """

        pass

    def add_to_toolbar(self, toolvar_item):
        """ Adds a tool bar item to a tool bar. """

        pass

    def can_add_to_menu(self, action):
        """ Returns True iff an action can be added to the menu. """

        return True

    def perform(self, action, event):
        """ Perform an action. """

        # fixme: We need a more formal event structure!
        event.widget = self
        event.context = self._context

        # fixme: the 'perform' method without taking an event is deprecated!
        args, varargs, varkw, defaults = getfullargspec(action.perform)

        # If the only argument is 'self' then this is the DEPRECATED
        # interface.
        if len(args) == 1:
            action.perform()

        else:
            action.perform(event)

        return

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _create_action_event(self, obj):
        """ Return a new action event for the specified object. """

        return ActionEvent(widget=self, context=obj)

    def _perform_default_action(self, action, obj):
        """ Perform the default action on the specified object. """

        action.perform(self._create_action_event(obj))

    def _popup_menu(self, menu_manager, obj, point):
        """ Popup the menu described by the menu manager. """

        # Create the actual menu control.
        menu = menu_manager.create_menu(self.control, self)
        if not menu.is_empty():
            # Show the menu. If an action is selected it will be performed
            # *before* this call returns.
            menu.show(*point)

            # This gives the actions in the menu manager a chance to cleanup
            # any event listeners etc.
            menu_manager.destroy()

        return



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A toolkit-specific tool bar manager that realizes itself in a tool palette
control.

- :attr:`~.ToolPaletteManager`
"""


# Import the toolkit specific version.
from pyface.toolkit import toolkit_object

ToolPaletteManager = toolkit_object(
    "action.tool_palette_manager:ToolPaletteManager"
)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


from collections.abc import Sequence
import importlib.util
import os
import shutil
import tempfile
import unittest

import pyface     # a package with images as package resources
from ..resource_manager import PyfaceResourceFactory
from ..resource_manager import ResourceManager

IMAGE_PATH = os.path.join(os.path.dirname(__file__), "images", "core.png")


class TestPyfaceResourceFactory(unittest.TestCase):
    def setUp(self):
        self.resource_factory = PyfaceResourceFactory()

    def test_image_from_file(self):
        image = self.resource_factory.image_from_file(IMAGE_PATH)

    def test_image_from_data(self):
        with open(IMAGE_PATH, "rb") as fp:
            data = fp.read()
        image = self.resource_factory.image_from_data(data)

    def test_locate_image(self):
        class ASequence(Sequence):
            def __init__(self, data):
                self.data = data

            def __getitem__(self, i):
                return self.data[i]

            def __len__(self):
                return len(self.data)

        sequence = ASequence([os.path.dirname(IMAGE_PATH)])

        resource_manager = ResourceManager()
        img_ref = resource_manager.locate_image("core.png", sequence)
        self.assertEqual(IMAGE_PATH, img_ref.filename)

    def test_locate_image_with_module(self):
        # ResourceManager should be able to find the images/close.png which
        # is included in pyface package data.
        resource_manager = ResourceManager()
        image_ref = resource_manager.locate_image("close.png", [pyface])
        self.assertGreater(len(image_ref.data), 0)

    def test_locate_image_with_module_missing_file(self):
        # The required image is not found, locate_image should return None.
        resource_manager = ResourceManager()
        image_ref = resource_manager.locate_image(
            "does_not_exist.png", [pyface]
        )
        self.assertIsNone(image_ref)

    def test_locate_image_with_name_being_dunder_main(self):
        # When a module is not a package, we will fall back to use __file__

        # given a module from which there is an image in the same folder
        with tempfile.TemporaryDirectory() as tmp_dir:
            shutil.copyfile(IMAGE_PATH, os.path.join(tmp_dir, "random.png"))
            # create an empty file for creating a module.
            py_filepath = os.path.join(tmp_dir, "tmp.py")
            with open(py_filepath, "w", encoding="utf-8"):
                pass
            spec = importlib.util.spec_from_file_location(
                "__main__", py_filepath
            )
            module = importlib.util.module_from_spec(spec)

            resource_manager = ResourceManager(
                resource_factory=PyfaceResourceFactory()
            )
            # when
            image_ref = resource_manager.load_image("random.png", [module])

            # then
            self.assertIsNotNone(image_ref)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


from traits.api import Any, Event, Property, provides, Str
from traits.api import Tuple


from pyface.i_window import IWindow, MWindow
from pyface.key_pressed_event import KeyPressedEvent
from .widget import Widget


@provides(IWindow)
class Window(MWindow, Widget):
    """ The toolkit specific implementation of a Window.  See the IWindow
    interface for the API documentation.
    """

    # 'IWindow' interface -------------------------------------------------#

    position = Property(Tuple)

    size = Property(Tuple)

    title = Str()

    # Events -----

    activated = Event()

    closed = Event()

    closing = Event()

    deactivated = Event()

    key_pressed = Event(KeyPressedEvent)

    opened = Event()

    opening = Event()

    # Private interface ----------------------------------------------------

    # Shadow trait for position.
    _position = Tuple((-1, -1))

    # Shadow trait for size.
    _size = Tuple((-1, -1))

    # ------------------------------------------------------------------------
    # 'IWindow' interface.
    # ------------------------------------------------------------------------

    def show(self, visible):
        pass

    # ------------------------------------------------------------------------
    # Protected 'IWindow' interface.
    # ------------------------------------------------------------------------

    def _add_event_listeners(self):
        pass

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _get_position(self):
        """ Property getter for position. """

        return self._position

    def _set_position(self, position):
        """ Property setter for position. """

        old = self._position
        self._position = position

        self.trait_property_changed("position", old, position)

    def _get_size(self):
        """ Property getter for size. """

        return self._size

    def _set_size(self, size):
        """ Property setter for size. """

        old = self._size
        self._size = size

        self.trait_property_changed("size", old, size)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traits.api import Interface


from pyface.tasks.i_editor_area_pane import IEditorAreaPane


class IAdvancedEditorAreaPane(IEditorAreaPane):
    """ A splitable central pane that contains tabbed editors.
    """

    # ------------------------------------------------------------------------
    # 'IAdvancedEditorAreaPane' interface.
    # ------------------------------------------------------------------------

    def get_layout(self):
        """ Returns a LayoutItem that reflects the current state of the editors.

        Because editors do not have IDs, they are identified by their index in
        the list of editors. For example, PaneItem(0) refers to the first
        editor.

        If there are no open editors, returns None.
        """

    def set_layout(self, layout):
        """ Applies a LayoutItem to the editors in the pane.

        The layout should have panes with IDs as described in ``get_layout()``.
        For example, if one wanted to open two editors side by side, with the
        first to the left of the right, something like this would appropriate::

            editor_area.edit(File('foo.py'))
            editor_area.edit(File('bar.py'))
            editor_area.set_layout(VSplitter(PaneItem(0),
                                             PaneItem(1)))

        Editors that are not included in the layout will be tabbified with other
        editors in an undefined manner.

        If the layout is None, this method is a no-op. Hence it is always safe
        to call this method with the results of ``get_layout()``.
        """



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


""" Enthought pyface package component
"""


import sys


import wx


from traits.api import HasTraits, Int, Property, provides, Tuple


from pyface.i_system_metrics import ISystemMetrics, MSystemMetrics


@provides(ISystemMetrics)
class SystemMetrics(MSystemMetrics, HasTraits):
    """ The toolkit specific implementation of a SystemMetrics.  See the
    ISystemMetrics interface for the API documentation.
    """

    # 'ISystemMetrics' interface -------------------------------------------

    screen_width = Property(Int)

    screen_height = Property(Int)

    dialog_background_color = Property(Tuple)

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _get_screen_width(self):
        return wx.SystemSettings.GetMetric(wx.SYS_SCREEN_X)

    def _get_screen_height(self):
        return wx.SystemSettings.GetMetric(wx.SYS_SCREEN_Y)

    def _get_dialog_background_color(self):
        if sys.platform == "darwin":
            # wx lies.
            color = wx.Colour(232, 232, 232)
        else:
            color = wx.SystemSettings.GetColour(wx.SYS_COLOUR_MENUBAR).Get()

        return (color[0] / 255.0, color[1] / 255.0, color[2] / 255.0)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Color conversion routines for the null toolkit.

This module provides a couple of utility methods to support the
pyface.color.Color class to_toolkit and from_toolkit methods.

For definiteness, the null toolkit uses tuples of RGBA values from 0 to 255
to represent colors.
"""

from pyface.color import channels_to_ints, ints_to_channels


def toolkit_color_to_rgba(color):
    """ Convert a hex tuple to an RGBA tuple.

    Parameters
    ----------
    color : tuple
        A tuple of integer values from 0 to 255 inclusive.

    Returns
    -------
    rgba : tuple
        A tuple of 4 floating point values between 0.0 and 1.0 inclusive.
    """
    return ints_to_channels(color)


def rgba_to_toolkit_color(rgba):
    """ Convert an RGBA tuple to a hex tuple.

    Parameters
    ----------
    color : tuple
        A tuple of integer values from 0 to 255 inclusive.

    Returns
    -------
    rgba : tuple
        A tuple of 4 floating point values between 0.0 and 1.0 inclusive.
    """
    return channels_to_ints(rgba)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" A UI that allows the user to choose a view. """


from pyface.workbench.api import IView, WorkbenchWindow
from traits.api import Any, HasTraits, Instance, List, Str
from traits.api import TraitError, Undefined
from traitsui.api import Item, TreeEditor, TreeNode, View
from traitsui.menu import Action  # fixme: Non-api import!


class Category(HasTraits):
    """ A view category. """

    # The name of the category.
    name = Str()

    # The views in the category.
    views = List()


class WorkbenchWindowTreeNode(TreeNode):
    """ A tree node for workbench windows that displays the window's views.

    The views are grouped by their category.

    """

    # 'TreeNode' interface -------------------------------------------------

    # List of object classes that the node applies to.
    node_for = [WorkbenchWindow]

    # ------------------------------------------------------------------------
    # 'TreeNode' interface.
    # ------------------------------------------------------------------------

    def get_children(self, object):
        """ Get the object's children. """

        # Collate the window's views into categories.
        categories_by_name = self._get_categories_by_name(object)

        categories = list(categories_by_name.values())
        categories.sort(key=lambda category: category.name)

        return categories

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _get_categories_by_name(self, window):
        """ Return a dictionary containing all categories keyed by name. """

        categories_by_name = {}
        for view in window.views:
            category = categories_by_name.get(view.category)
            if category is None:
                category = Category(name=view.category)
                categories_by_name[view.category] = category

            category.views.append(view)

        return categories_by_name


class IViewTreeNode(TreeNode):
    """ A tree node for objects that implement the 'IView' interface.

    This node does *not* recognise objects that can be *adapted* to the 'IView'
    interface, only those that actually implement it. If we wanted to allow
    for adaptation we would have to work out a way for the rest of the
    'TreeNode' code to access the adapter, not the original object. We could,
    of course override every method, but that seems a little, errr, tedious.
    We could probably do with something like in the Pyface tree where there
    is a method that returns the actual object that we want to manipulate.

    """

    def is_node_for(self, obj):
        """ Returns whether this is the node that handles a specified object.

        """

        # By checking for 'is obj' here, we are *not* allowing adaptation (if
        # we were allowing adaptation it would be 'is not None'). See the class
        # doc string for details.
        return IView(obj, Undefined) is obj

    def get_icon(self, obj, is_expanded):
        """ Returns the icon for a specified object. """

        if obj.image is not None:
            icon = obj.image

        else:
            # fixme: A bit of magic here! Is there a better way to say 'use
            # the default leaf icon'?
            icon = "<item>"

        return icon


class ViewChooser(HasTraits):
    """ Allow the user to choose a view.

    This implementation shows views in a tree grouped by category.

    """

    # The window that contains the views to choose from.
    window = Instance("pyface.workbench.api.WorkbenchWindow")

    # The currently selected tree item (at any point in time this might be
    # either None, a view category, or a view).
    selected = Any()

    # The selected view (None if the selected item is not a view).
    view = Instance(IView)

    # Traits UI views -----------------------------------------------------#

    traits_ui_view = View(
        Item(
            name="window",
            editor=TreeEditor(
                nodes=[
                    WorkbenchWindowTreeNode(
                        auto_open=True,
                        label="=Views",
                        rename=False,
                        copy=False,
                        delete=False,
                        insert=False,
                        menu=None,
                    ),
                    TreeNode(
                        node_for=[Category],
                        auto_open=True,
                        children="views",
                        label="name",
                        rename=False,
                        copy=False,
                        delete=False,
                        insert=False,
                        menu=None,
                    ),
                    IViewTreeNode(
                        auto_open=False,
                        label="name",
                        rename=False,
                        copy=False,
                        delete=False,
                        insert=False,
                        menu=None,
                    ),
                ],
                editable=False,
                hide_root=True,
                selected="selected",
                show_icons=True,
            ),
            show_label=False,
        ),
        buttons=[Action(name="OK", enabled_when="view is not None"), "Cancel"],
        resizable=True,
        style="custom",
        title="Show View",
        width=0.2,
        height=0.4,
    )

    # ------------------------------------------------------------------------
    # 'ViewChooser' interface.
    # ------------------------------------------------------------------------

    def _selected_changed(self, old, new):
        """ Static trait change handler. """

        # If the assignment fails then the selected object does *not* implement
        # the 'IView' interface.
        try:
            self.view = new

        except TraitError:
            self.view = None

        return



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The wx specific implementations the action manager internal classes.
"""


from inspect import getfullargspec


import wx


from traits.api import Any, Bool, HasTraits


from pyface.action.action_event import ActionEvent


_STYLE_TO_KIND_MAP = {
    "push": wx.ITEM_NORMAL,
    "radio": wx.ITEM_RADIO,
    "toggle": wx.ITEM_CHECK,
    "widget": None,
}


class _MenuItem(HasTraits):
    """ A menu item representation of an action item. """

    # '_MenuItem' interface ------------------------------------------------

    # Is the item checked?
    checked = Bool(False)

    # A controller object we delegate taking actions through (if any).
    controller = Any()

    # Is the item enabled?
    enabled = Bool(True)

    # Is the item visible?
    visible = Bool(True)

    # The radio group we are part of (None if the menu item is not part of such
    # a group).
    group = Any()

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(self, parent, menu, item, controller):
        """ Creates a new menu item for an action item. """

        self.item = item

        # Create an appropriate menu item depending on the style of the action.
        #
        # N.B. Don't try to use -1 as the Id for the menu item... wx does not
        # like it!
        action = item.action
        label = action.name
        kind = _STYLE_TO_KIND_MAP[action.style]
        longtip = action.description or action.tooltip

        if action.style == "widget":
            raise NotImplementedError(
                "WxPython does not support widgets in menus"
            )

        if len(action.accelerator) > 0:
            label = label + "\t" + action.accelerator

        # This just helps with debugging when people forget to specify a name
        # for their action (without this wx just barfs which is not very
        # helpful!).
        if len(label) == 0:
            label = item.action.__class__.__name__

        if getattr(action, "menu_role", False):
            if action.menu_role == "About":
                self.control_id = wx.ID_ABOUT
            elif action.menu_role == "Preferences":
                self.control_id = wx.ID_PREFERENCES
            elif action.menu_role == "Quit":
                self.control_id = wx.ID_EXIT
        else:
            self.control_id = wx.ID_ANY
        self.control = wx.MenuItem(menu, self.control_id, label, longtip, kind)

        # If the action has an image then display it.
        if action.image is not None:
            try:
                self.control.SetBitmap(action.image.create_bitmap())
            except Exception:
                # Some platforms don't allow radio buttons to have
                # bitmaps, so just ignore the exception if it happens
                pass

        menu.Append(self.control)
        menu.menu_items.append(self)

        # Set the initial enabled/disabled state of the action.
        self.control.Enable(action.enabled and action.visible)

        # Set the initial checked state.
        if action.style in ["radio", "toggle"]:
            self.control.Check(action.checked)

        # Wire it up...create an ugly flag since some platforms dont skip the
        # event when we thought they would
        self._skip_menu_event = False
        parent.Bind(wx.EVT_MENU, self._on_menu, self.control)

        # Listen for trait changes on the action (so that we can update its
        # enabled/disabled/checked state etc).
        action.observe(self._on_action_enabled_changed, "enabled")
        action.observe(self._on_action_visible_changed, "visible")
        action.observe(self._on_action_checked_changed, "checked")
        action.observe(self._on_action_name_changed, "name")
        action.observe(self._on_action_image_changed, "image")

        if controller is not None:
            self.controller = controller
            controller.add_to_menu(self)

    def dispose(self):
        action = self.item.action
        action.observe(self._on_action_enabled_changed, "enabled", remove=True)
        action.observe(self._on_action_visible_changed, "visible", remove=True)
        action.observe(self._on_action_checked_changed, "checked", remove=True)
        action.observe(self._on_action_name_changed, "name", remove=True)
        action.observe(self._on_action_image_changed, "image", remove=True)

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    # Trait event handlers -------------------------------------------------

    def _enabled_changed(self):
        """ Called when our 'enabled' trait is changed. """

        self.control.Enable(self.enabled and self.visible)

    def _visible_changed(self):
        """ Called when our 'visible' trait is changed. """

        self.control.Enable(self.visible and self.enabled)

    def _checked_changed(self):
        """ Called when our 'checked' trait is changed. """

        if self.item.action.style == "radio":
            # fixme: Not sure why this is even here, we had to guard it to
            # make it work? Must take a look at svn blame!
            # FIXME v3: Note that menu_checked() doesn't seem to exist, so we
            # comment it out and do the following instead.
            # if self.group is not None:
            #    self.group.menu_checked(self)

            # If we're turning this one on, then we need to turn all the others
            # off.  But if we're turning this one off, don't worry about the
            # others.
            if self.checked:
                for item in self.item.parent.items:
                    if item is not self.item:
                        item.action.checked = False

        self.control.Check(self.checked)

    def _on_action_enabled_changed(self, event):
        """ Called when the enabled trait is changed on an action. """
        action = event.object
        self.control.Enable(action.enabled and action.visible)

    def _on_action_visible_changed(self, event):
        """ Called when the visible trait is changed on an action. """
        action = event.object
        self.control.Enable(action.visible and action.enabled)

    def _on_action_checked_changed(self, event):
        """ Called when the checked trait is changed on an action. """
        action = event.object
        if self.item.action.style == "radio":
            # fixme: Not sure why this is even here, we had to guard it to
            # make it work? Must take a look at svn blame!
            # FIXME v3: Note that menu_checked() doesn't seem to exist, so we
            # comment it out and do the following instead.
            # if self.group is not None:
            #    self.group.menu_checked(self)

            # If we're turning this one on, then we need to turn all the others
            # off.  But if we're turning this one off, don't worry about the
            # others.
            if action.checked:
                for item in self.item.parent.items:
                    if item is not self.item:
                        item.action.checked = False

        # This will *not* emit a menu event because of this ugly flag
        self._skip_menu_event = True
        self.control.Check(action.checked)
        self._skip_menu_event = False

    def _on_action_name_changed(self, event):
        """ Called when the name trait is changed on an action. """
        action = event.object
        label = action.name
        if len(action.accelerator) > 0:
            label = label + "\t" + action.accelerator
        self.control.SetText(label)

    def _on_action_image_changed(self, event):
        """ Called when the name trait is changed on an action. """
        action = event.object
        if self.control is not None:
            self.control.SetIcon(action.image.create_icon())

        return

    # wx event handlers ----------------------------------------------------

    def _on_menu(self, event):
        """ Called when the menu item is clicked. """

        # if the ugly flag is set, do not perform the menu event
        if self._skip_menu_event:
            return

        action = self.item.action
        action_event = ActionEvent()

        is_checkable = action.style in ["radio", "toggle"]

        # Perform the action!
        if self.controller is not None:
            if is_checkable:
                # fixme: There is a difference here between having a controller
                # and not in that in this case we do not set the checked state
                # of the action! This is confusing if you start off without a
                # controller and then set one as the action now behaves
                # differently!
                self.checked = self.control.IsChecked() == 1

            # Most of the time, action's do no care about the event (it
            # contains information about the time the event occurred etc), so
            # we only pass it if the perform method requires it. This is also
            # useful as Traits UI controllers *never* require the event.
            argspec = getfullargspec(self.controller.perform)

            # If the only arguments are 'self' and 'action' then don't pass
            # the event!
            if len(argspec.args) == 2:
                self.controller.perform(action)

            else:
                self.controller.perform(action, action_event)

        else:
            if is_checkable:
                action.checked = self.control.IsChecked() == 1

            # Most of the time, action's do no care about the event (it
            # contains information about the time the event occurred etc), so
            # we only pass it if the perform method requires it.
            argspec = getfullargspec(action.perform)

            # If the only argument is 'self' then don't pass the event!
            if len(argspec.args) == 1:
                action.perform()

            else:
                action.perform(action_event)


class _Tool(HasTraits):
    """ A tool bar tool representation of an action item. """

    # '_Tool' interface ----------------------------------------------------

    # Is the item checked?
    checked = Bool(False)

    # A controller object we delegate taking actions through (if any).
    controller = Any()

    # Is the item enabled?
    enabled = Bool(True)

    # Is the item visible?
    visible = Bool(True)

    # The radio group we are part of (None if the tool is not part of such a
    # group).
    group = Any()

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(
        self, parent, tool_bar, image_cache, item, controller, show_labels
    ):
        """ Creates a new tool bar tool for an action item. """

        self.item = item
        self.tool_bar = tool_bar

        # Create an appropriate tool depending on the style of the action.
        action = self.item.action
        label = action.name

        # Tool bar tools never have '...' at the end!
        if label.endswith("..."):
            label = label[:-3]

        # And they never contain shortcuts.
        label = label.replace("&", "")

        # If the action has an image then convert it to a bitmap (as required
        # by the toolbar).
        if action.image is not None:
            image = action.image.create_image(
                self.tool_bar.GetToolBitmapSize()
            )
            path = action.image.absolute_path
            bmp = image_cache.get_bitmap(path)

        else:
            from pyface.api import ImageResource

            image = ImageResource("image_not_found")
            bmp = image.create_bitmap()

        kind = _STYLE_TO_KIND_MAP[action.style]
        tooltip = action.tooltip
        longtip = action.description

        if not show_labels:
            label = ""

        else:
            self.tool_bar.SetSize((-1, 50))

        if action.style == "widget":
            widget = action.create_control(self.tool_bar)
            self.control = tool_bar.AddControl(widget, label)
            self.control_id = self.control.GetId()
        else:
            self.control = tool_bar.AddTool(
                wx.ID_ANY,
                label,
                bmp,
                wx.NullBitmap,
                kind,
                tooltip,
                longtip,
                None,
            )
            self.control_id = self.control.GetId()

            # Set the initial checked state.
            tool_bar.ToggleTool(self.control_id, action.checked)

            if hasattr(tool_bar, "ShowTool"):
                # Set the initial enabled/disabled state of the action.
                tool_bar.EnableTool(self.control_id, action.enabled)

                # Set the initial visibility
                tool_bar.ShowTool(self.control_id, action.visible)
            else:
                # Set the initial enabled/disabled state of the action.
                tool_bar.EnableTool(
                    self.control_id, action.enabled and action.visible
                )

            # Wire it up.
            parent.Bind(wx.EVT_TOOL, self._on_tool, self.control)

        # Listen for trait changes on the action (so that we can update its
        # enabled/disabled/checked state etc).
        action.observe(self._on_action_enabled_changed, "enabled")
        action.observe(self._on_action_visible_changed, "visible")
        action.observe(self._on_action_checked_changed, "checked")

        if controller is not None:
            self.controller = controller
            controller.add_to_toolbar(self)

    def dispose(self):
        action = self.item.action
        action.observe(self._on_action_enabled_changed, "enabled", remove=True)
        action.observe(self._on_action_visible_changed, "visible", remove=True)
        action.observe(self._on_action_checked_changed, "checked", remove=True)

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    # Trait event handlers -------------------------------------------------

    def _enabled_changed(self):
        """ Called when our 'enabled' trait is changed. """

        if hasattr(self.tool_bar, "ShowTool"):
            self.tool_bar.EnableTool(self.control_id, self.enabled)
        else:
            self.tool_bar.EnableTool(
                self.control_id, self.enabled and self.visible
            )

    def _visible_changed(self):
        """ Called when our 'visible' trait is changed. """

        if hasattr(self.tool_bar, "ShowTool"):
            self.tool_bar.ShowTool(self.control_id, self.visible)
        else:
            self.tool_bar.EnableTool(
                self.control_id, self.enabled and self.visible
            )

    def _checked_changed(self):
        """ Called when our 'checked' trait is changed. """

        if self.item.action.style == "radio":
            # FIXME v3: Note that toolbar_checked() doesn't seem to exist, so
            # we comment it out and do the following instead.
            # self.group.toolbar_checked(self)

            # If we're turning this one on, then we need to turn all the others
            # off.  But if we're turning this one off, don't worry about the
            # others.
            if self.checked:
                for item in self.item.parent.items:
                    if item is not self.item:
                        item.action.checked = False

        self.tool_bar.ToggleTool(self.control_id, self.checked)

    def _on_action_enabled_changed(self, event):
        """ Called when the enabled trait is changed on an action. """
        action = event.object
        if hasattr(self.tool_bar, "ShowTool"):
            self.tool_bar.EnableTool(self.control_id, action.enabled)
        else:
            self.tool_bar.EnableTool(
                self.control_id, action.enabled and action.visible
            )

    def _on_action_visible_changed(self, event):
        """ Called when the visible trait is changed on an action. """
        action = event.object
        if hasattr(self.tool_bar, "ShowTool"):
            self.tool_bar.ShowTool(self.control_id, action.visible)
        else:
            self.tool_bar.EnableTool(
                self.control_id, self.enabled and action.visible
            )

    def _on_action_checked_changed(self, event):
        """ Called when the checked trait is changed on an action. """
        action = event.object
        if action.style == "radio":
            # If we're turning this one on, then we need to turn all the others
            # off.  But if we're turning this one off, don't worry about the
            # others.
            if event.new:
                for item in self.item.parent.items:
                    if item is not self.item:
                        item.action.checked = False

        # This will *not* emit a tool event.
        self.tool_bar.ToggleTool(self.control_id, event.new)

        return

    # wx event handlers ----------------------------------------------------

    def _on_tool(self, event):
        """ Called when the tool bar tool is clicked. """

        action = self.item.action
        action_event = ActionEvent()

        is_checkable = action.style == "radio" or action.style == "check"

        # Perform the action!
        if self.controller is not None:
            # fixme: There is a difference here between having a controller
            # and not in that in this case we do not set the checked state
            # of the action! This is confusing if you start off without a
            # controller and then set one as the action now behaves
            # differently!
            self.checked = self.tool_bar.GetToolState(self.control_id) == 1

            # Most of the time, action's do no care about the event (it
            # contains information about the time the event occurred etc), so
            # we only pass it if the perform method requires it. This is also
            # useful as Traits UI controllers *never* require the event.
            argspec = getfullargspec(self.controller.perform)

            # If the only arguments are 'self' and 'action' then don't pass
            # the event!
            if len(argspec.args) == 2:
                self.controller.perform(action)

            else:
                self.controller.perform(action, action_event)

        else:
            action.checked = self.tool_bar.GetToolState(self.control_id) == 1

            # Most of the time, action's do no care about the event (it
            # contains information about the time the event occurred etc), so
            # we only pass it if the perform method requires it.
            argspec = getfullargspec(action.perform)

            # If the only argument is 'self' then don't pass the event!
            if len(argspec.args) == 1:
                action.perform()

            else:
                action.perform(action_event)


class _PaletteTool(HasTraits):
    """ A tool palette representation of an action item. """

    # '_PaletteTool' interface ---------------------------------------------

    # The radio group we are part of (None if the tool is not part of such a
    # group).
    group = Any()

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(self, tool_palette, image_cache, item, show_labels):
        """ Creates a new tool palette tool for an action item. """

        self.item = item
        self.tool_palette = tool_palette

        action = self.item.action
        label = action.name

        if action.style == "widget":
            raise NotImplementedError(
                "WxPython does not support widgets in palettes"
            )

        # Tool palette tools never have '...' at the end.
        if label.endswith("..."):
            label = label[:-3]

        # And they never contain shortcuts.
        label = label.replace("&", "")

        image = action.image.create_image()
        path = action.image.absolute_path
        bmp = image_cache.get_bitmap(path)

        kind = action.style
        tooltip = action.tooltip
        longtip = action.description

        if not show_labels:
            label = ""

        # Add the tool to the tool palette.
        self.tool_id = tool_palette.add_tool(
            label, bmp, kind, tooltip, longtip
        )
        tool_palette.toggle_tool(self.tool_id, action.checked)
        tool_palette.enable_tool(self.tool_id, action.enabled)
        tool_palette.on_tool_event(self.tool_id, self._on_tool)

        # Listen to the trait changes on the action (so that we can update its
        # enabled/disabled/checked state etc).
        action.observe(self._on_action_enabled_changed, "enabled")
        action.observe(self._on_action_checked_changed, "checked")

        return

    def dispose(self):
        action = self.item.action
        action.observe(self._on_action_enabled_changed, "enabled", remove=True)
        action.observe(self._on_action_checked_changed, "checked", remove=True)

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    # Trait event handlers -------------------------------------------------

    def _on_action_enabled_changed(self, event):
        """ Called when the enabled trait is changed on an action. """
        action = event.object
        self.tool_palette.enable_tool(self.tool_id, action.enabled)

    def _on_action_checked_changed(self, event):
        """ Called when the checked trait is changed on an action. """
        action = event.object
        if action.style == "radio":
            # If we're turning this one on, then we need to turn all the others
            # off.  But if we're turning this one off, don't worry about the
            # others.
            if event.new:
                for item in self.item.parent.items:
                    if item is not self.item:
                        item.action.checked = False

        # This will *not* emit a tool event.
        self.tool_palette.toggle_tool(self.tool_id, event.new)

        return

    # Tool palette event handlers -----------------------------------------#

    def _on_tool(self, event):
        """ Called when the tool palette button is clicked. """

        action = self.item.action
        action_event = ActionEvent()

        is_checkable = action.style == "radio" or action.style == "check"

        # Perform the action!
        action.checked = self.tool_palette.get_tool_state(self.tool_id) == 1
        action.perform(action_event)

        return



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Base class for all tree label providers. """


from .label_provider import LabelProvider


class TreeLabelProvider(LabelProvider):
    """ Base class for all tree label providers.

    By default an element has no label image, and 'str' is used to generate its
    label text.

    """

    # ------------------------------------------------------------------------
    # 'LabelProvider' interface.
    # ------------------------------------------------------------------------

    def set_text(self, viewer, element, text):
        """ Sets the text representation of a node.

        Returns True if setting the text succeeded, otherwise False.

        """

        return len(text.strip()) > 0

    # ------------------------------------------------------------------------
    # 'TreeLabelProvider' interface.
    # ------------------------------------------------------------------------

    def get_drag_value(self, viewer, element):
        """ Get the value that is dragged for an element.

        By default the drag value is the element itself.

        """

        return element

    def is_collapsible(self, viewer, element):
        """ Returns True is the element is collapsible, otherwise False. """

        return True

    def is_expandable(self, viewer, node):
        """ Returns True is the node is expandanble, otherwise False. """

        return True



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traits.api import HasTraits, Instance


from .task_pane import TaskPane


class TraitsTaskPane(TaskPane):
    """ A TaskPane that displays a Traits UI View.
    """

    # TraitsTaskPane interface ---------------------------------------------

    #: The model object to view. If not specified, the pane is used instead.
    model = Instance(HasTraits)

    #: The UI object associated with the Traits view, if it has been
    #: constructed.
    ui = Instance("traitsui.ui.UI")

    # ------------------------------------------------------------------------
    # 'HasTraits' interface.
    # ------------------------------------------------------------------------

    def trait_context(self):
        """ Use the model object for the Traits UI context, if appropriate.
        """
        if self.model:
            return {"object": self.model, "pane": self}
        return super(TraitsTaskPane, self).trait_context()

    # ------------------------------------------------------------------------
    # 'ITaskPane' interface.
    # ------------------------------------------------------------------------

    def create(self, parent):
        """ Create and set the toolkit-specific control that represents the
            pane.
        """
        self.ui = self.edit_traits(kind="subpanel", parent=parent)
        self.control = self.ui.control

    def destroy(self):
        """ Destroy the toolkit-specific control that represents the pane.
        """
        self.ui.dispose()
        self.control = self.ui = None



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


""" Enthought pyface package component
"""


from logging import DEBUG


import wx
import wx.adv


from traits.api import Any, Bool, Font, Instance, Int, provides
from traits.api import Tuple, Str


from pyface.i_splash_screen import ISplashScreen, MSplashScreen
from pyface.image_resource import ImageResource
from pyface.wx.util.font_helper import new_font_like
from .window import Window


@provides(ISplashScreen)
class SplashScreen(MSplashScreen, Window):
    """ The toolkit specific implementation of a SplashScreen.  See the
    ISplashScreen interface for the API documentation.
    """

    # 'ISplashScreen' interface --------------------------------------------

    image = Instance(ImageResource, ImageResource("splash"))

    log_level = Int(DEBUG)

    show_log_messages = Bool(True)

    text = Str()

    text_color = Any()

    text_font = Any()

    text_location = Tuple(5, 5)

    # ------------------------------------------------------------------------
    # Protected 'IWidget' interface.
    # ------------------------------------------------------------------------

    def _create_control(self, parent):
        # Get the splash screen image.
        image = self.image.create_image()

        splash_screen = wx.adv.SplashScreen(
            # The bitmap to display on the splash screen.
            image.ConvertToBitmap(),
            # Splash Style.
            wx.adv.SPLASH_NO_TIMEOUT | wx.adv.SPLASH_CENTRE_ON_SCREEN,
            # Timeout in milliseconds (we don't currently timeout!).
            0,
            # The parent of the splash screen.
            parent,
            # wx Id.
            -1,
            # Window style.
            style=wx.SIMPLE_BORDER | wx.FRAME_NO_TASKBAR,
        )

        # By default we create a font slightly bigger and slightly more italic
        # than the normal system font ;^)  The font is used inside the event
        # handler for 'EVT_PAINT'.
        self._wx_default_text_font = new_font_like(
            wx.NORMAL_FONT,
            point_size=wx.NORMAL_FONT.GetPointSize() + 1,
            style=wx.ITALIC,
        )

        # This allows us to write status text on the splash screen.
        splash_screen.Bind(wx.EVT_PAINT, self._on_paint)

        return splash_screen

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _text_changed(self):
        """ Called when the splash screen text has been changed. """

        # Passing 'False' to 'Refresh' means "do not erase the background".
        if self.control is not None:
            self.control.Refresh(False)
            self.control.Update()
        wx.GetApp().Yield(True)

    def _on_paint(self, event):
        """ Called when the splash window is being repainted. """

        if self.control is not None:
            # Get the window that the splash image is drawn in.
            window = self.control  # .GetSplashWindow()

            dc = wx.PaintDC(window)

            if self.text_font is None:
                text_font = self._wx_default_text_font
            else:
                text_font = self.text_font

            dc.SetFont(text_font)

            if self.text_color is None:
                text_color = "black"
            else:
                text_color = self.text_color

            dc.SetTextForeground(text_color)

            x, y = self.text_location
            dc.DrawText(self.text, x, y)

        # Let the normal wx paint handling do its stuff.
        event.Skip()



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Abstract base class for tree content providers. """


from .content_provider import ContentProvider


class TreeContentProvider(ContentProvider):
    """ Abstract base class for tree content providers.

    Tree content providers are used by (surprise, surprise) tree viewers!

    """

    # ------------------------------------------------------------------------
    # 'ContentProvider' interface.
    # ------------------------------------------------------------------------

    def get_elements(self, element):
        """ Returns a list of the elements to display in a viewer.

        Returns a list of elements to display in a viewer when its (ie. the
        viewer's) input is set to the given element.

        The returned list should not be modified by the viewer.

        """

        return get_children(element)

    # ------------------------------------------------------------------------
    # 'TreeContentProvider' interface.
    # ------------------------------------------------------------------------

    def get_parent(self, element):
        """ Returns the parent of an element.

        Returns None if the element either has no parent (ie. it is the root of
        the tree), or if the parent cannot be computed.

        """

        return None

    def get_children(self, element):
        """ Returns the children of an element. """

        raise NotImplementedError()

    def has_children(self, element):
        """ Returns True iff the element has children, otherwise False. """

        raise NotImplementedError()



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


import unittest

from ..action import Action
from ..action_event import ActionEvent


class TestAction(unittest.TestCase):
    def test_default_id(self):
        action = Action(name="Test")
        self.assertEqual(action.id, "Test")

    def test_id(self):
        action = Action(name="Test", id="test")
        self.assertEqual(action.id, "test")

    def test_perform(self):
        # test whether function is called by updating list
        # XXX should really use mock
        memo = []

        def perform():
            memo.append("called")

        action = Action(name="Test", on_perform=perform)
        event = ActionEvent()
        action.perform(event)
        self.assertEqual(memo, ["called"])

    def test_perform_none(self):
        action = Action(name="Test")
        event = ActionEvent()
        # does nothing, but shouldn't error
        action.perform(event)

    def test_destroy(self):
        action = Action(name="Test")
        # does nothing, but shouldn't error
        action.destroy()

    def test_widget_action(self):
        # test whether function is called by updating list
        # XXX should really use mock
        memo = []

        def control_factory(parent, action):
            memo.append((parent, action))

        action = Action(
            name="Dummy", style="widget", control_factory=control_factory
        )
        parent = None
        action.create_control(parent)

        self.assertEqual(memo, [(parent, action)])



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A wizard controller that can be chained with others. """


from traits.api import Instance


from .i_wizard_controller import IWizardController
from .wizard_controller import WizardController


class ChainedWizardController(WizardController):
    """ A wizard controller that can be chained with others. """

    # 'ChainedWizardController' interface ---------------------------------#

    # The next chained wizard controller.
    next_controller = Instance(IWizardController)

    # ------------------------------------------------------------------------
    # 'IWizardController' interface.
    # ------------------------------------------------------------------------

    def get_next_page(self, page):
        """ Returns the next page. """

        next_page = None

        if page in self._pages:
            if page is not self._pages[-1]:
                index = self._pages.index(page)
                next_page = self._pages[index + 1]

            else:
                if self.next_controller is not None:
                    next_page = self.next_controller.get_first_page()

        else:
            if self.next_controller is not None:
                next_page = self.next_controller.get_next_page(page)

        return next_page

    def get_previous_page(self, page):
        """ Returns the previous page. """

        if page in self._pages:
            index = self._pages.index(page)
            previous_page = self._pages[index - 1]

        else:
            if self.next_controller is not None:
                if self.next_controller.is_first_page(page):
                    previous_page = self._pages[-1]

                else:
                    previous_page = self.next_controller.get_previous_page(
                        page
                    )

            else:
                previous_page = None

        return previous_page

    def is_first_page(self, page):
        """ Is the page the first page? """

        return page is self._pages[0]

    def is_last_page(self, page):
        """ Is the page the last page? """

        if page in self._pages:
            # If page is not this controller's last page, then it cannot be
            # *the* last page.
            if not page is self._pages[-1]:
                is_last = False

            # Otherwise, it is *the* last page if this controller has no next
            # controller or the next controller has no pages.
            else:
                if self.next_controller is None:
                    is_last = True

                else:
                    is_last = self.next_controller.is_last_page(page)

        else:
            if self.next_controller is not None:
                is_last = self.next_controller.is_last_page(page)

            elif len(self._pages) > 0:
                is_last = False

            else:
                is_last = True

        return is_last

    def dispose_pages(self):
        """ Dispose the wizard's pages. """

        for page in self._pages:
            page.dispose_page()

        if self.next_controller is not None:
            self.next_controller.dispose_pages()

        return

    # ------------------------------------------------------------------------
    # 'ChainedWizardController' interface.
    # ------------------------------------------------------------------------

    def _get_pages(self):
        """ Returns the pages in the wizard. """

        pages = self._pages[:]

        if self.next_controller is not None:
            pages.extend(self.next_controller.pages)

        return pages

    def _set_pages(self, pages):
        """ Sets the pages in the wizard. """

        self._pages = pages

        return

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _update(self):
        """ Checks the completion status of the controller. """

        # The entire wizard is complete when ALL pages are complete.
        for page in self._pages:
            if not page.complete:
                self.complete = False
                break

        else:
            if self.next_controller is not None:
                # fixme: This is a abstraction leak point, since _update is not
                # part of the wizard_controller interface!
                self.next_controller._update()
                self.complete = self.next_controller.complete

            else:
                self.complete = True

        return

    # Trait event handlers -------------------------------------------------

    # Static ----

    def _current_page_changed(self, old, new):
        """ Called when the current page is changed. """

        if old is not None:
            old.observe(
                self._on_page_complete, "complete", remove=True
            )

        if new is not None:
            new.observe(self._on_page_complete, "complete")

        if self.next_controller is not None:
            self.next_controller.current_page = new

        self._update()

    def _next_controller_changed(self, old, new):
        """ Called when the next controller is changed. """

        if old is not None:
            old.observe(
                self._on_controller_complete, "complete", remove=True
            )

        if new is not None:
            new.observe(self._on_controller_complete, "complete")

        self._update()

        return

    # Dynamic ----

    def _on_controller_complete(self, event):
        """ Called when the next controller's complete state changes. """

        self._update()

    def _on_page_complete(self, event):
        """ Called when the current page is complete. """

        self._update()

        return



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


""" Enthought pyface package component
"""


import os
import tempfile
from io import BytesIO


import wx


from pyface.resource.api import ResourceFactory

from traits.api import Undefined


class PyfaceResourceFactory(ResourceFactory):
    """ The implementation of a shared resource manager. """

    # ------------------------------------------------------------------------
    # 'ResourceFactory' toolkit interface.
    # ------------------------------------------------------------------------

    def image_from_file(self, filename):
        """ Creates an image from the data in the specified filename. """

        # N.B 'wx.BITMAP_TYPE_ANY' tells wxPython to attempt to autodetect the
        # --- image format.
        return wx.Image(filename, wx.BITMAP_TYPE_ANY)

    def image_from_data(self, data, filename=None):
        """ Creates an image from the specified data. """
        return wx.Image(BytesIO(data))

        handle = None
        if filename is None:
            # If there is currently no way in wx to create an image from data,
            # we have write it out to a temporary file and then read it back in:
            handle, filename = tempfile.mkstemp()

        # Write it out...
        tf = open(filename, "wb")
        tf.write(data)
        tf.close()

        # ... and read it back in!  Lovely 8^()
        image = wx.Image(filename, wx.BITMAP_TYPE_ANY)

        # Remove the temporary file.
        if handle is not None:
            os.close(handle)
            os.unlink(filename)

        return image



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
# (C) Copyright 2008 Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms described in the PyQt GPL exception also apply


import logging


from pyface.qt import QtCore, QtGui


from traits.api import Instance, observe


from pyface.message_dialog import error
from pyface.workbench.i_workbench_window_layout import MWorkbenchWindowLayout
from .split_tab_widget import SplitTabWidget


# Logging.
logger = logging.getLogger(__name__)


# For mapping positions relative to the editor area.
_EDIT_AREA_MAP = {
    "left": QtCore.Qt.LeftDockWidgetArea,
    "right": QtCore.Qt.RightDockWidgetArea,
    "top": QtCore.Qt.TopDockWidgetArea,
    "bottom": QtCore.Qt.BottomDockWidgetArea,
}

# For mapping positions relative to another view.
_VIEW_AREA_MAP = {
    "left": (QtCore.Qt.Horizontal, True),
    "right": (QtCore.Qt.Horizontal, False),
    "top": (QtCore.Qt.Vertical, True),
    "bottom": (QtCore.Qt.Vertical, False),
}


class WorkbenchWindowLayout(MWorkbenchWindowLayout):
    """ The Qt4 implementation of the workbench window layout interface.

    See the 'IWorkbenchWindowLayout' interface for the API documentation.

    """

    # Private interface ----------------------------------------------------

    # The widget that provides the editor area.  We keep (and use) this
    # separate reference because we can't always assume that it has been set to
    # be the main window's central widget.
    _qt4_editor_area = Instance(SplitTabWidget)

    # ------------------------------------------------------------------------
    # 'IWorkbenchWindowLayout' interface.
    # ------------------------------------------------------------------------

    def activate_editor(self, editor):
        if editor.control is not None:
            editor.control.show()
            self._qt4_editor_area.setCurrentWidget(editor.control)
            editor.set_focus()

        return editor

    def activate_view(self, view):
        # FIXME v3: This probably doesn't work as expected.
        view.control.raise_()
        view.set_focus()

        return view

    def add_editor(self, editor, title):
        if editor is None:
            return None

        try:
            self._qt4_editor_area.addTab(
                self._qt4_get_editor_control(editor), title
            )

            if editor._loading_on_open:
                self._qt4_editor_tab_spinner(editor, "", True)
        except Exception:
            logger.exception("error creating editor control [%s]", editor.id)

        return editor

    def add_view(self, view, position=None, relative_to=None, size=(-1, -1)):
        if view is None:
            return None

        try:
            self._qt4_add_view(view, position, relative_to, size)
            view.visible = True
        except Exception:
            logger.exception("error creating view control [%s]", view.id)

            # Even though we caught the exception, it sometimes happens that
            # the view's control has been created as a child of the application
            # window (or maybe even the dock control).  We should destroy the
            # control to avoid bad UI effects.
            view.destroy_control()

            # Additionally, display an error message to the user.
            error(
                self.window.control,
                "Unable to add view [%s]" % view.id,
                "Workbench Plugin Error",
            )

        return view

    def close_editor(self, editor):
        if editor.control is not None:
            editor.control.close()

        return editor

    def close_view(self, view):
        self.hide_view(view)

        return view

    def close(self):
        # Don't fire signals for editors that have destroyed their controls.
        self._qt4_editor_area.editor_has_focus.disconnect(
            self._qt4_editor_focus
        )

        self._qt4_editor_area.clear()

        # Delete all dock widgets.
        for v in self.window.views:
            if self.contains_view(v):
                self._qt4_delete_view_dock_widget(v)

    def create_initial_layout(self, parent):
        self._qt4_editor_area = editor_area = SplitTabWidget(parent)

        editor_area.editor_has_focus.connect(self._qt4_editor_focus)

        # We are interested in focus changes but we get them from the editor
        # area rather than qApp to allow the editor area to restrict them when
        # needed.
        editor_area.focus_changed.connect(self._qt4_view_focus_changed)

        editor_area.tabTextChanged.connect(self._qt4_editor_title_changed)
        editor_area.new_window_request.connect(self._qt4_new_window_request)
        editor_area.tab_close_request.connect(self._qt4_tab_close_request)
        editor_area.tab_window_changed.connect(self._qt4_tab_window_changed)

        return editor_area

    def contains_view(self, view):
        return hasattr(view, "_qt4_dock")

    def hide_editor_area(self):
        self._qt4_editor_area.hide()

    def hide_view(self, view):
        view._qt4_dock.hide()
        view.visible = False

        return view

    def refresh(self):
        # Nothing to do.
        pass

    def reset_editors(self):
        self._qt4_editor_area.setCurrentIndex(0)

    def reset_views(self):
        # Qt doesn't provide information about the order of dock widgets in a
        # dock area.
        pass

    def show_editor_area(self):
        self._qt4_editor_area.show()

    def show_view(self, view):
        view._qt4_dock.show()
        view.visible = True

    # Methods for saving and restoring the layout -------------------------#

    def get_view_memento(self):
        # Get the IDs of the views in the main window.  This information is
        # also in the QMainWindow state, but that is opaque.
        view_ids = [v.id for v in self.window.views if self.contains_view(v)]

        # Everything else is provided by QMainWindow.
        state = self.window.control.saveState()

        return (0, (view_ids, state))

    def set_view_memento(self, memento):
        version, mdata = memento

        # There has only ever been version 0 so far so check with an assert.
        assert version == 0

        # Now we know the structure of the memento we can "parse" it.
        view_ids, state = mdata

        # Get a list of all views that have dock widgets and mark them.
        dock_views = [v for v in self.window.views if self.contains_view(v)]

        for v in dock_views:
            v._qt4_gone = True

        # Create a dock window for all views that had one last time.
        for v in self.window.views:
            # Make sure this is in a known state.
            v.visible = False

            for vid in view_ids:
                if vid == v.id:
                    # Create the dock widget if needed and make sure that it is
                    # invisible so that it matches the state of the visible
                    # trait.  Things will all come right when the main window
                    # state is restored below.
                    self._qt4_create_view_dock_widget(v).setVisible(False)

                    if v in dock_views:
                        delattr(v, "_qt4_gone")

                    break

        # Remove any remain unused dock widgets.
        for v in dock_views:
            try:
                delattr(v, "_qt4_gone")
            except AttributeError:
                pass
            else:
                self._qt4_delete_view_dock_widget(v)

        # Restore the state.  This will update the view's visible trait through
        # the dock window's toggle action.
        self.window.control.restoreState(state)

    def get_editor_memento(self):
        # Get the layout of the editors.
        editor_layout = self._qt4_editor_area.saveState()

        # Get a memento for each editor that describes its contents.
        editor_references = self._get_editor_references()

        return (0, (editor_layout, editor_references))

    def set_editor_memento(self, memento):
        version, mdata = memento

        # There has only ever been version 0 so far so check with an assert.
        assert version == 0

        # Now we know the structure of the memento we can "parse" it.
        editor_layout, editor_references = mdata

        def resolve_id(id):
            # Get the memento for the editor contents (if any).
            editor_memento = editor_references.get(id)

            if editor_memento is None:
                return None

            # Create the restored editor.
            editor = self.window.editor_manager.set_editor_memento(
                editor_memento
            )
            if editor is None:
                return None

            # Save the editor.
            self.window.editors.append(editor)

            # Create the control if needed and return it.
            return self._qt4_get_editor_control(editor)

        self._qt4_editor_area.restoreState(editor_layout, resolve_id)

    def get_toolkit_memento(self):
        return (0, {"geometry": self.window.control.saveGeometry()})

    def set_toolkit_memento(self, memento):
        if hasattr(memento, "toolkit_data"):
            data = memento.toolkit_data
            if isinstance(data, tuple) and len(data) == 2:
                version, datadict = data
                if version == 0:
                    geometry = datadict.pop("geometry", None)
                    if geometry is not None:
                        self.window.control.restoreGeometry(geometry)

    def is_editor_area_visible(self):
        return self._qt4_editor_area.isVisible()

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _qt4_editor_focus(self, new):
        """ Handle an editor getting the focus. """

        for editor in self.window.editors:
            control = editor.control
            editor.has_focus = control is new or (
                control is not None and new in control.children()
            )

    def _qt4_editor_title_changed(self, control, title):
        """ Handle the title being changed """
        for editor in self.window.editors:
            if editor.control == control:
                editor.name = str(title)

    def _qt4_editor_tab_spinner(self, event):
        editor = event.object

        # Do we need to do this verification?
        tw, tidx = self._qt4_editor_area._tab_widget(editor.control)

        if event.new:
            tw.show_button(tidx)
        else:
            tw.hide_button(tidx)

        if not event.new and not editor == self.window.active_editor:
            self._qt4_editor_area.setTabTextColor(
                editor.control, QtCore.Qt.red
            )

    @observe("window:active_editor")
    def _qt4_active_editor_changed(self, event):
        """ Handle change of active editor """
        # Reset tab title to foreground color
        editor = event.new 
        if editor is not None:
            self._qt4_editor_area.setTabTextColor(editor.control)

    def _qt4_view_focus_changed(self, old, new):
        """ Handle the change of focus for a view. """

        focus_part = None

        if new is not None:
            # Handle focus changes to views.
            for view in self.window.views:
                if view.control is not None and view.control.isAncestorOf(new):
                    view.has_focus = True
                    focus_part = view
                    break

        if old is not None:
            # Handle focus changes from views.
            for view in self.window.views:
                if (
                    view is not focus_part
                    and view.control is not None
                    and view.control.isAncestorOf(old)
                ):
                    view.has_focus = False
                    break

    def _qt4_new_window_request(self, pos, control):
        """ Handle a tab tear-out request from the splitter widget. """

        editor = self._qt4_remove_editor_with_control(control)
        kind = self.window.editor_manager.get_editor_kind(editor)

        window = self.window.workbench.create_window()
        window.open()
        window.add_editor(editor)
        window.editor_manager.add_editor(editor, kind)
        window.position = (pos.x(), pos.y())
        window.size = self.window.size
        window.activate_editor(editor)
        editor.window = window

    def _qt4_tab_close_request(self, control):
        """ Handle a tabCloseRequest from the splitter widget. """

        for editor in self.window.editors:
            if editor.control == control:
                editor.close()
                break

    def _qt4_tab_window_changed(self, control):
        """ Handle a tab drag to a different WorkbenchWindow. """

        editor = self._qt4_remove_editor_with_control(control)
        kind = self.window.editor_manager.get_editor_kind(editor)

        while not control.isWindow():
            control = control.parent()
        for window in self.window.workbench.windows:
            if window.control == control:
                window.editors.append(editor)
                window.editor_manager.add_editor(editor, kind)
                window.layout._qt4_get_editor_control(editor)
                window.activate_editor(editor)
                editor.window = window
                break

    def _qt4_remove_editor_with_control(self, control):
        """ Finds the editor associated with 'control' and removes it. Returns
            the editor, or None if no editor was found.
        """
        for editor in self.window.editors:
            if editor.control == control:
                self.editor_closing = editor
                control.removeEventFilter(self._qt4_mon)
                self.editor_closed = editor

                # Make sure that focus events get fired if this editor is
                # subsequently added to another window.
                editor.has_focus = False

                return editor

    def _qt4_get_editor_control(self, editor):
        """ Create the editor control if it hasn't already been done. """

        if editor.control is None:
            self.editor_opening = editor

            # We must provide a parent (because TraitsUI checks for it when
            # deciding what sort of panel to create) but it can't be the editor
            # area (because it will be automatically added to the base
            # QSplitter).
            editor.control = editor.create_control(self.window.control)
            editor.control.setObjectName(editor.id)

            editor.observe(self._qt4_editor_tab_spinner, "_loading")

            self.editor_opened = editor

        def on_name_changed(event):
            editor = event.object
            self._qt4_editor_area.setWidgetTitle(editor.control, editor.name)

        editor.observe(on_name_changed, "name")

        self._qt4_monitor(editor.control)

        return editor.control

    def _qt4_add_view(self, view, position, relative_to, size):
        """ Add a view. """

        # If no specific position is specified then use the view's default
        # position.
        if position is None:
            position = view.position

        dw = self._qt4_create_view_dock_widget(view, size)
        mw = self.window.control

        try:
            rel_dw = relative_to._qt4_dock
        except AttributeError:
            rel_dw = None

        if rel_dw is None:
            # If we are trying to add a view with a non-existent item, then
            # just default to the left of the editor area.
            if position == "with":
                position = "left"

            # Position the view relative to the editor area.
            try:
                dwa = _EDIT_AREA_MAP[position]
            except KeyError:
                raise ValueError("unknown view position: %s" % position)

            mw.addDockWidget(dwa, dw)
        elif position == "with":
            # FIXME v3: The Qt documentation says that the second should be
            # placed above the first, but it always seems to be underneath (ie.
            # hidden) which is not what the user is expecting.
            mw.tabifyDockWidget(rel_dw, dw)
        else:
            try:
                orient, swap = _VIEW_AREA_MAP[position]
            except KeyError:
                raise ValueError("unknown view position: %s" % position)

            mw.splitDockWidget(rel_dw, dw, orient)

            # The Qt documentation implies that the layout direction can be
            # used to position the new dock widget relative to the existing one
            # but I could only get the button positions to change.  Instead we
            # move things around afterwards if required.
            if swap:
                mw.removeDockWidget(rel_dw)
                mw.splitDockWidget(dw, rel_dw, orient)
                rel_dw.show()

    def _qt4_create_view_dock_widget(self, view, size=(-1, -1)):
        """ Create a dock widget that wraps a view. """

        # See if it has already been created.
        try:
            dw = view._qt4_dock
        except AttributeError:
            dw = QtGui.QDockWidget(view.name, self.window.control)
            dw.setWidget(_ViewContainer(size, self.window.control))
            dw.setObjectName(view.id)
            dw.toggleViewAction().toggled.connect(
                self._qt4_handle_dock_visibility
            )
            dw.visibilityChanged.connect(self._qt4_handle_dock_visibility)

            # Save the dock window.
            view._qt4_dock = dw

            def on_name_changed(event):
                view._qt4_dock.setWindowTitle(view.name)

            view.observe(on_name_changed, "name")

        # Make sure the view control exists.
        if view.control is None:
            # Make sure that the view knows which window it is in.
            view.window = self.window

            try:
                view.control = view.create_control(dw.widget())
            except:
                # Tidy up if the view couldn't be created.
                delattr(view, "_qt4_dock")
                self.window.control.removeDockWidget(dw)
                dw.deleteLater()
                del dw
                raise

        dw.widget().setCentralWidget(view.control)

        return dw

    def _qt4_delete_view_dock_widget(self, view):
        """ Delete a view's dock widget. """

        dw = view._qt4_dock

        # Disassociate the view from the dock.
        if view.control is not None:
            view.control.setParent(None)

        delattr(view, "_qt4_dock")

        # Delete the dock (and the view container).
        self.window.control.removeDockWidget(dw)
        dw.deleteLater()

    def _qt4_handle_dock_visibility(self, checked):
        """ Handle the visibility of a dock window changing. """

        # Find the dock window by its toggle action.
        for v in self.window.views:
            try:
                dw = v._qt4_dock
            except AttributeError:
                continue

            sender = dw.sender()
            if sender is dw.toggleViewAction() or sender in dw.children():
                # Toggling the action or pressing the close button on
                # the view
                v.visible = checked

    def _qt4_monitor(self, control):
        """ Install an event filter for a view or editor control to keep an eye
        on certain events.
        """

        # Create the monitoring object if needed.
        try:
            mon = self._qt4_mon
        except AttributeError:
            mon = self._qt4_mon = _Monitor(self)

        control.installEventFilter(mon)


class _Monitor(QtCore.QObject):
    """ This class monitors a view or editor control. """

    def __init__(self, layout):
        QtCore.QObject.__init__(self, layout.window.control)

        self._layout = layout

    def eventFilter(self, obj, e):
        if isinstance(e, QtGui.QCloseEvent):
            for editor in self._layout.window.editors:
                if editor.control is obj:
                    self._layout.editor_closing = editor
                    editor.destroy_control()
                    self._layout.editor_closed = editor

                    break

        return False


class _ViewContainer(QtGui.QMainWindow):
    """ This class is a container for a view that allows an initial size
    (specified as a tuple) to be set.
    """

    def __init__(self, size, main_window):
        """ Initialise the object. """

        QtGui.QMainWindow.__init__(self)

        # Save the size and main window.
        self._width, self._height = size
        self._main_window = main_window

    def sizeHint(self):
        """ Reimplemented to return the initial size or the view's current
        size.
        """

        sh = self.centralWidget().sizeHint()

        if self._width > 0:
            if self._width > 1:
                w = self._width
            else:
                w = self._main_window.width() * self._width

            sh.setWidth(int(w))

        if self._height > 0:
            if self._height > 1:
                h = self._height
            else:
                h = self._main_window.height() * self._height

            sh.setHeight(int(h))

        return sh

    def showEvent(self, e):
        """ Reimplemented to use the view's current size once shown. """

        self._width = self._height = -1

        QtGui.QMainWindow.showEvent(self, e)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""

API for the ``pyface.tasks.action`` subpackage.

ActionSchemas and aliases
-------------------------

- :class:`~.ActionSchema`
- :class:`~.GroupSchema`
- :class:`~.MenuSchema`
- :class:`~.MenuBarSchema`
- :class:`~.ToolBarSchema`
- :attr:`~.SGroup`
- :attr:`~.SMenu`
- :attr:`~.SMenuBar`
- :attr:`~.SToolBar`

Schema Addition
---------------

- :class:`~.SchemaAddition`

Tasks-specific Action Controller
--------------------------------

- :class:`~.TaskActionController`

Tasks-specific Action Manager factory
-------------------------------------

- :class:`~.TaskActionManagerBuilder`

Tasks-specific Actions
----------------------

- :class:`~.CentralPaneAction`
- :class:`~.DockPaneAction`
- :class:`~.EditorAction`
- :class:`~.TaskAction`
- :class:`~.TaskWindowAction`
- :class:`~.TasksApplicationAction`

Useful Tasks Actions and Groups
-------------------------------

- :class:`~.DockPaneToggleGroup`
- :class:`~.TaskToggleGroup`
- :class:`~.TaskWindowToggleGroup`
- :class:`~.CreateTaskWindowAction`

"""

from .dock_pane_toggle_group import DockPaneToggleGroup
from .schema import (
    ActionSchema,
    GroupSchema,
    MenuSchema,
    MenuBarSchema,
    ToolBarSchema,
    SGroup,
    SMenu,
    SMenuBar,
    SToolBar,
)
from .schema_addition import SchemaAddition
from .task_action import (
    CentralPaneAction,
    DockPaneAction,
    EditorAction,
    TaskAction,
    TaskWindowAction,
)
from .task_action_controller import TaskActionController
from .task_action_manager_builder import TaskActionManagerBuilder
from .task_toggle_group import TaskToggleGroup
from .task_window_toggle_group import TaskWindowToggleGroup
from .tasks_application_action import (
    CreateTaskWindowAction,
    TasksApplicationAction,
)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


""" The wx specific implementation of a menu manager.
"""


import wx


from traits.api import Str, Bool


from pyface.action.action_manager import ActionManager
from pyface.action.action_manager_item import ActionManagerItem
from pyface.action.group import Group


class MenuManager(ActionManager, ActionManagerItem):
    """ A menu manager realizes itself in a menu control.

    This could be a sub-menu or a context (popup) menu.
    """

    # 'MenuManager' interface ---------------------------------------------#

    # The menu manager's name (if the manager is a sub-menu, this is what its
    # label will be).
    name = Str()

    # Does the menu require a separator before the menu item name?
    separator = Bool(True)

    # ------------------------------------------------------------------------
    # 'MenuManager' interface.
    # ------------------------------------------------------------------------

    def create_menu(self, parent, controller=None):
        """ Creates a menu representation of the manager. """

        # If a controller is required it can either be set as a trait on the
        # menu manager (the trait is part of the 'ActionManager' API), or
        # passed in here (if one is passed in here it takes precedence over the
        # trait).
        if controller is None:
            controller = self.controller

        return _Menu(self, parent, controller)

    # ------------------------------------------------------------------------
    # 'ActionManagerItem' interface.
    # ------------------------------------------------------------------------

    def add_to_menu(self, parent, menu, controller):
        """ Adds the item to a menu. """

        sub = self.create_menu(parent, controller)
        id = sub.GetId()

        # fixme: Nasty hack to allow enabling/disabling of menus.
        sub._id = id
        sub._menu = menu

        menu.Append(id, self.name, sub)

    def add_to_toolbar(self, parent, tool_bar, image_cache, controller):
        """ Adds the item to a tool bar. """

        raise ValueError("Cannot add a menu manager to a toolbar.")


class _Menu(wx.Menu):
    """ The toolkit-specific menu control. """

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(self, manager, parent, controller):
        """ Creates a new tree. """

        # Base class constructor.
        wx.Menu.__init__(self)

        # The parent of the menu.
        self._parent = parent

        # The manager that the menu is a view of.
        self._manager = manager

        # The controller.
        self._controller = controller

        # List of menu items
        self.menu_items = []

        # Create the menu structure.
        self.refresh()

        # Listen to the manager being updated.
        self._manager.observe(self.refresh, "changed")
        self._manager.observe(self._on_enabled_changed, "enabled")

        return

    def dispose(self):
        self._manager.observe(self.refresh, "changed", remove=True)
        self._manager.observe(self._on_enabled_changed, "enabled", remove=True)
        # Removes event listeners from downstream menu items
        self.clear()

    # ------------------------------------------------------------------------
    # '_Menu' interface.
    # ------------------------------------------------------------------------

    def clear(self):
        """ Clears the items from the menu. """

        for item in self.GetMenuItems():
            if item.GetSubMenu() is not None:
                item.GetSubMenu().clear()
            self.Delete(item.GetId())

        for item in self.menu_items:
            item.dispose()

        self.menu_items = []

    def is_empty(self):
        """ Is the menu empty? """

        return self.GetMenuItemCount() == 0

    def refresh(self, event=None):
        """ Ensures that the menu reflects the state of the manager. """

        self.clear()

        manager = self._manager
        parent = self._parent

        previous_non_empty_group = None
        for group in manager.groups:
            previous_non_empty_group = self._add_group(
                parent, group, previous_non_empty_group
            )

    def show(self, x=None, y=None):
        """ Show the menu at the specified location. """

        if x is None or y is None:
            self._parent.PopupMenu(self)
        else:
            self._parent.PopupMenu(self, x, y)

        return

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _on_enabled_changed(self, event):
        """ Dynamic trait change handler. """

        # fixme: Nasty hack to allow enabling/disabling of menus.
        #
        # We cannot currently (AFAIK) disable menus on the menu bar. Hence
        # we don't give them an '_id'...

        if hasattr(self, "_id"):
            self._menu.Enable(self._id, event.new)

    def _add_group(self, parent, group, previous_non_empty_group=None):
        """ Adds a group to a menu. """

        if len(group.items) > 0:
            # Is a separator required?
            if previous_non_empty_group is not None and group.separator:
                self.AppendSeparator()

            # Create actions and sub-menus for each contribution item in
            # the group.
            for item in group.items:
                if isinstance(item, Group):
                    if len(item.items) > 0:
                        self._add_group(parent, item, previous_non_empty_group)

                        if (
                            previous_non_empty_group is not None
                            and previous_non_empty_group.separator
                            and item.separator
                        ):
                            self.AppendSeparator()

                        previous_non_empty_group = item

                else:
                    if isinstance(item, MenuManager):
                        if item.separator:
                            self.AppendSeparator()
                        previous_non_empty_group = item
                    item.add_to_menu(parent, self, self._controller)

            previous_non_empty_group = group

        return previous_non_empty_group



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A widget for editing Python code. """


from traits.api import Bool, Event, HasTraits, Interface, Str


from pyface.key_pressed_event import KeyPressedEvent


class IPythonEditor(Interface):
    """ A widget for editing Python code. """

    # 'IPythonEditor' interface --------------------------------------------

    #: Has the file in the editor been modified?
    dirty = Bool(False)

    #: The pathname of the file being edited.
    path = Str()

    #: Should line numbers be shown in the margin?
    show_line_numbers = Bool(True)

    # Events ----

    #: The contents of the editor has changed.
    changed = Event()

    #: A key has been pressed.
    key_pressed = Event(KeyPressedEvent)

    # ------------------------------------------------------------------------
    # 'IPythonEditor' interface.
    # ------------------------------------------------------------------------

    def load(self, path=None):
        """ Loads the contents of the editor.

        Parameters
        ----------
        path : str or None
            The path to the file to load.
        """

    def save(self, path=None):
        """ Saves the contents of the editor.

        Parameters
        ----------
        path : str or None
            The path to the file to save.
        """

    # FIXME v3: This is very dependent on the underlying implementation.
    def set_style(self, n, fore, back):
        """ Set the foreground and background colors for a particular style and
        set the font and size to default values.
        """

    def select_line(self, lineno):
        """ Selects the specified line.

        Parameters
        ----------
        lineno : int
            The line number to select.
        """


class MPythonEditor(HasTraits):
    """ The mixin class that contains common code for toolkit specific
    implementations of the IPythonEditor interface.

    Implements: _changed_path()
    """

    def _changed_path(self):
        """ Called when the path to the file is changed. """

        if self.control is not None:
            self.load()



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A thin visual divider. """


import wx


class Divider(wx.StaticLine):
    """ A thin visual divider. """

    def __init__(self, parent, id, **kw):
        """ Creates a divider. """

        # Base-class constructor.
        wx.StaticLine.__init__(self, parent, id, style=wx.LI_HORIZONTAL, **kw)

        # Create the widget!
        self._create_widget()

        return

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _create_widget(self):
        """ Creates the widget. """

        self.SetSize((1, 1))

        return



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" An action that renames a user perspective. """


from .user_perspective_action import UserPerspectiveAction
from .user_perspective_name import UserPerspectiveName


class RenameUserPerspectiveAction(UserPerspectiveAction):
    """ An action that renames a user perspective. """

    # 'Action' interface ---------------------------------------------------

    # The action's unique identifier (may be None).
    id = "pyface.workbench.action.rename_user_perspective_action"

    # The action's name (displayed on menus/tool bar tools etc).
    name = "Rename Perspective..."

    # ------------------------------------------------------------------------
    # 'Action' interface.
    # ------------------------------------------------------------------------

    def perform(self, event):
        """ Perform the action. """

        window = event.window
        manager = window.workbench.user_perspective_manager

        # Get the new name.
        upn = UserPerspectiveName(name=window.active_perspective.name)
        if upn.edit_traits(view="rename_view").result:
            manager.rename(window.active_perspective, upn.name.strip())

        return



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


from contextlib import contextmanager
from io import BytesIO
import logging
from pickle import dumps, load, loads

import wx

from traits.api import provides
from pyface.i_clipboard import IClipboard, BaseClipboard


logger = logging.getLogger(__name__)

# Data formats
PythonObjectFormat = wx.DataFormat("PythonObject")
TextFormat = wx.DataFormat(wx.DF_TEXT)
FileFormat = wx.DataFormat(wx.DF_FILENAME)

# Shortcuts
cb = wx.TheClipboard


@contextmanager
def _ensure_clipboard():
    """ Ensure use of X11 clipboard rather than primary selection on X11.

    X11 allows pasting from either the clipboard or the primary selection.
    This context manager ensures that the clipboard is always used for Pyface,
    no matter what the wider application state is currently using.

    On non-X11 platforms this does nothing.
    """
    using_primary = cb.IsUsingPrimarySelection()
    if using_primary:
        cb.UsePrimarySelection(False)
        try:
            yield
        finally:
            cb.UsePrimarySelection(True)
    else:
        yield


@contextmanager
def _close_clipboard(flush=False):
    """ Ensures clipboard is closed and (optionally) flushed.

    Parameters
    ----------
    flush : bool
        Whether or not to flush the clipboard.  Should be true when setting
        data to the clipboard.
    """
    try:
        yield
    finally:
        cb.Close()
        if flush:
            cb.Flush()


@provides(IClipboard)
class Clipboard(BaseClipboard):
    """ WxPython implementation of the IClipboard interface.

    Python object data is transmitted as bytes consisting of the pickled class
    object followed by the corresponding pickled instance object.  This means
    that copy/paste of Python objects may not work unless compatible Python
    libraries are available at the pasting location.
    """

    # ---------------------------------------------------------------------------
    #  'data' property methods:
    # ---------------------------------------------------------------------------

    def _get_has_data(self):
        result = False
        with _ensure_clipboard():
            if cb.Open():
                with _close_clipboard():
                    result = (
                        cb.IsSupported(TextFormat)
                        or cb.IsSupported(FileFormat)
                        or cb.IsSupported(PythonObjectFormat)
                    )
        return result

    # ---------------------------------------------------------------------------
    #  'object_data' property methods:
    # ---------------------------------------------------------------------------

    def _get_object_data(self):
        result = None
        with _ensure_clipboard():
            if cb.Open():
                with _close_clipboard():
                    if cb.IsSupported(PythonObjectFormat):
                        cdo = wx.CustomDataObject(PythonObjectFormat)
                        if cb.GetData(cdo):
                            file = BytesIO(cdo.GetData())
                            _ = load(file)
                            result = load(file)
        return result

    def _set_object_data(self, data):
        with _ensure_clipboard():
            if cb.Open():
                with _close_clipboard(flush=True):
                    cdo = wx.CustomDataObject(PythonObjectFormat)
                    cdo.SetData(dumps(data.__class__) + dumps(data))
                    # fixme: There seem to be cases where the '-1' value creates
                    # pickles that can't be unpickled (e.g. some TraitDictObject's)
                    # cdo.SetData(dumps(data, -1))
                    cb.SetData(cdo)

    def _get_has_object_data(self):
        return self._has_this_data(PythonObjectFormat)

    def _get_object_type(self):
        result = ""
        with _ensure_clipboard():
            if cb.Open():
                with _close_clipboard():
                    if cb.IsSupported(PythonObjectFormat):
                        cdo = wx.CustomDataObject(PythonObjectFormat)
                        if cb.GetData(cdo):
                            try:
                                # We may not be able to load the required class:
                                result = loads(cdo.GetData())
                            except Exception:
                                logger.exception("Cannot load data from clipboard.")
        return result

    # ---------------------------------------------------------------------------
    #  'text_data' property methods:
    # ---------------------------------------------------------------------------

    def _get_text_data(self):
        result = ""
        with _ensure_clipboard():
            if cb.Open():
                with _close_clipboard():
                    if cb.IsSupported(TextFormat):
                        tdo = wx.TextDataObject()
                        if cb.GetData(tdo):
                            result = tdo.GetText()
            return result

    def _set_text_data(self, data):
        with _ensure_clipboard():
            if cb.Open():
                with _close_clipboard(flush=True):
                    cb.SetData(wx.TextDataObject(str(data)))

    def _get_has_text_data(self):
        return self._has_this_data(TextFormat)

    # ---------------------------------------------------------------------------
    #  'file_data' property methods:
    # ---------------------------------------------------------------------------

    def _get_file_data(self):
        with _ensure_clipboard():
            result = []
            if cb.Open():
                with _close_clipboard():
                    if cb.IsSupported(FileFormat):
                        tfo = wx.FileDataObject()
                        if cb.GetData(tfo):
                            result = tfo.GetFilenames()
            return result

    def _set_file_data(self, data):
        with _ensure_clipboard():
            if cb.Open():
                with _close_clipboard(flush=True):
                    tfo = wx.FileDataObject()
                    if isinstance(data, str):
                        tfo.AddFile(data)
                    else:
                        for filename in data:
                            tfo.AddFile(filename)
                    cb.SetData(tfo)

    def _get_has_file_data(self):
        return self._has_this_data(FileFormat)

    # ---------------------------------------------------------------------------
    #  Private helper methods:
    # ---------------------------------------------------------------------------

    def _has_this_data(self, format):
        result = False
        with _ensure_clipboard():
            if cb.Open():
                with _close_clipboard():
                    result = cb.IsSupported(format)

        return result



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
# (C) Copyright 2007 Riverbank Computing Limited
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


import sys


from pyface.qt import QtGui


from pyface.action.api import MenuBarManager, StatusBarManager
from pyface.action.api import ToolBarManager
from traits.api import Instance, List, observe, provides, Str


from pyface.i_application_window import IApplicationWindow, MApplicationWindow
from pyface.image_resource import ImageResource
from .window import Window


@provides(IApplicationWindow)
class ApplicationWindow(MApplicationWindow, Window):
    """ The toolkit specific implementation of an ApplicationWindow.  See the
    IApplicationWindow interface for the API documentation.
    """

    # 'IApplicationWindow' interface ---------------------------------------

    #: The icon to display in the application window title bar.
    icon = Instance(ImageResource)

    #: The menu bar manager for the window.
    menu_bar_manager = Instance(MenuBarManager)

    #: The status bar manager for the window.
    status_bar_manager = Instance(StatusBarManager)

    #: DEPRECATED: The tool bar manager for the window.
    tool_bar_manager = Instance(ToolBarManager)

    #: The collection of tool bar managers for the window.
    tool_bar_managers = List(ToolBarManager)

    # 'IWindow' interface -------------------------------------------------#

    #: The window title.
    title = Str("Pyface")

    # ------------------------------------------------------------------------
    # Protected 'IApplicationWindow' interface.
    # ------------------------------------------------------------------------

    def _create_contents(self, parent):
        panel = QtGui.QWidget(parent)

        palette = QtGui.QPalette(panel.palette())
        palette.setColor(QtGui.QPalette.Window, QtGui.QColor("blue"))
        panel.setPalette(palette)
        panel.setAutoFillBackground(True)

        return panel

    def _create_menu_bar(self, parent):
        if self.menu_bar_manager is not None:
            menu_bar = self.menu_bar_manager.create_menu_bar(parent)
            self.control.setMenuBar(menu_bar)

    def _create_status_bar(self, parent):
        if self.status_bar_manager is not None:
            status_bar = self.status_bar_manager.create_status_bar(parent)

            # QMainWindow automatically makes the status bar visible, but we
            # have delegated this responsibility to the status bar manager.
            self.control.setStatusBar(status_bar)
            status_bar.setVisible(self.status_bar_manager.visible)

    def _create_tool_bar(self, parent):
        tool_bar_managers = self._get_tool_bar_managers()
        visible = self.control.isVisible()
        for tool_bar_manager in tool_bar_managers:
            # Add the tool bar and make sure it is visible.
            tool_bar = tool_bar_manager.create_tool_bar(parent)
            self.control.addToolBar(tool_bar)
            tool_bar.show()

            # Make sure that the tool bar has a name so its state can be saved.
            if len(tool_bar.objectName()) == 0:
                tool_bar.setObjectName(tool_bar_manager.name)

        if sys.platform == "darwin":
            # Work around bug in Qt on OS X where creating a tool bar with a
            # QMainWindow parent hides the window. See
            # http://bugreports.qt.nokia.com/browse/QTBUG-5069 for more info.
            self.control.setVisible(visible)

    def _set_window_icon(self):
        if self.icon is None:
            icon = ImageResource("application.png")
        else:
            icon = self.icon
        if self.control is not None:
            self.control.setWindowIcon(icon.create_icon())

    # ------------------------------------------------------------------------
    # 'Window' interface.
    # ------------------------------------------------------------------------

    def _size_default(self):
        """ Trait initialiser. """

        return (800, 600)

    # ------------------------------------------------------------------------
    # Protected 'IWidget' interface.
    # ------------------------------------------------------------------------

    def _create(self):
        super(ApplicationWindow, self)._create()

        contents = self._create_contents(self.control)
        self.control.setCentralWidget(contents)

        self._create_trim_widgets(self.control)

    def _create_control(self, parent):
        control = super(ApplicationWindow, self)._create_control(parent)
        control.setObjectName("ApplicationWindow")

        control.setAnimated(False)
        control.setDockNestingEnabled(True)

        return control

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _get_tool_bar_managers(self):
        """ Return all tool bar managers specified for the window. """

        # fixme: V3 remove the old-style single toolbar option!
        if self.tool_bar_manager is not None:
            tool_bar_managers = [self.tool_bar_manager]

        else:
            tool_bar_managers = self.tool_bar_managers

        return tool_bar_managers

    # Trait change handlers ------------------------------------------------

    # QMainWindow takes ownership of the menu bar and the status bar upon
    # assignment. For this reason, it is unnecessary to delete the old controls
    # in the following two handlers.

    def _menu_bar_manager_changed(self):
        if self.control is not None:
            self._create_menu_bar(self.control)

    def _status_bar_manager_changed(self, old, new):
        if self.control is not None:
            if old is not None:
                old.destroy_status_bar()
            self._create_status_bar(self.control)

    @observe("tool_bar_manager, tool_bar_managers.items")
    def _update_tool_bar_managers(self, event):
        if self.control is not None:
            # Remove the old toolbars.
            for child in self.control.children():
                if isinstance(child, QtGui.QToolBar):
                    self.control.removeToolBar(child)
                    child.deleteLater()

            # Add the new toolbars.
            self._create_tool_bar(self.control)

    def _icon_changed(self):
        self._set_window_icon()



# (C) Copyright 2007-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# ------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
# Thanks for using Enthought open source!
#
# Author: Riverbank Computing Limited
# Description: <Enthought undo package component>
# ------------------------------------------------------------------------------

# Local imports.
from .abstract_command_stack_action import AbstractCommandStackAction


class UndoAction(AbstractCommandStackAction):
    """ An action that undos the last command of the active command stack. """

    ###########################################################################
    # 'Action' interface.
    ###########################################################################

    def perform(self, event):
        """ Perform the action. """

        self.undo_manager.undo()

    ###########################################################################
    # 'AbstractUndoAction' interface.
    ###########################################################################

    def _update_action(self):
        """ Update the state of the action. """

        name = self.undo_manager.undo_name

        if name:
            name = "&Undo " + name
            self.enabled = True
        else:
            name = "&Undo"
            self.enabled = False

        self.name = name
